#
# wbadmin.py
#
# MKBackup frontend for wbadmin
#
# (c) alain.spineux@gmail.com  


import os, sys, subprocess, time, platform
import cStringIO, zipfile, gzip
from datetime import datetime, timedelta
import xml.dom.minidom

import win32wnet
import win32netcon 

from windows import WindowsErrorDecode, windows_list_dir

from archiver import *

import cron

def getvalue(parentnode, nodename):
      return parentnode.getElementsByTagName(nodename)[0].childNodes[0].data

# ===========================================================================
#
# ===========================================================================
class Wbadmin(Archiver):

    name='wbadmin'
    exe='wbadmin.exe'
    
    types=dict(vsscopy='vsscopy', vssfull='vssfull', copy='copy') # copy is deprecated 
    
    # -----------------------------------------------------------------------
    def __init__(self):
        Archiver.__init__(self)
        
        self.selection=None
        self.verify=None
        self.restricted=None
        self.program_exe=None
        self.type=None
        self.target=None
        self.user=None
        self.password=None
        self.options=[]

        systemroot=os.environ.get('SystemRoot', os.environ.get('windir', 'C:\\windows'))
        self.wbadmin_bin=os.path.join(systemroot, 'system32', 'wbadmin.exe')
        self.wevtutil_bin=os.path.join(systemroot, 'system32', 'wevtutil.exe')
        # check if we are on a x64 running a 32bit MKBackup
        if platform.machine()=='x86' and os.path.exists(os.path.join(systemroot, 'sysnative', 'wbadmin.exe')):
                # yes we are
                self.wbadmin_bin=os.path.join(systemroot, 'sysnative', 'wbadmin.exe')
                self.wevtutil_bin=os.path.join(systemroot, 'sysnative', 'wevtutil.exe')

        self.machine=platform.node()

    # -----------------------------------------------------------------------
    def load(self, job, manager):
        
        log=manager.log
        now=manager.now

        errors, warnings, extra={}, {}, ''

        #
        # check job config
        #

        self.destination=job.get('destination', None)
        if self.destination:
            self.destination=self.destination.replace('\n','')
            try:
                self.destination=Destinations(self.destination, self)
            except (DestinationSyntaxError, cron.CronException), e:
                errors['destination']='syntax error: %s' % (str(e), )

        self.include=job.get('include', None)
        if self.include and not self.destination:
            errors['include']='destination required'

        self.user=job.get('user', None)
        if self.user and not self.destination:
            errors['user']='destination required'
            
        self.password=job.get('password', None)
        self.register_password(self.password)
            
        if self.user and not self.destination:
            errors['password']='destination required'

        if (self.user and not self.password) or (not self.user and self.password):
            errors['password']='login and password must be set together'

        options=job.get('options', '')
        try:
            options=quoted_string_list(options)
        except ValueError:
            errors['options']='not a valid quoted string list'
        else:
            self.options=options

        self.wbadmin_bin=job.get('wbadmin_bin', self.wbadmin_bin)
        self.wevtutil_bin=job.get('wevtutil_bin', self.wevtutil_bin)
        self.machine=job.get('machine', self.machine)

        extra+='wbadmin_bin=%s\n' % self.wbadmin_bin
        extra+='wevtutil_bin=%s\n' % self.wevtutil_bin
        
        if not errors:
            if not self.destination:
                self.type, self.target='default', None
            else:
                self.type, self.target=self.destination.match(now, self.night_shift)
    
                if self.type!=None and self.type!='none':            
                    # check the target
                    if not os.path.isdir(self.target):
                        try:
                            os.makedirs(self.target)
                        except WindowsError, e:
                            unc=os.path.splitunc(self.target)[0]
                            if unc and self.user and self.password:
                                try:
                                    win32wnet.WNetAddConnection2(win32netcon.RESOURCETYPE_DISK, None, param_encode(unc, manager.default_encoding), None, param_encode(self.user, manager.default_encoding), param_encode(self.password, manager.default_encoding))
                                    try:
                                        if not os.path.isdir(self.target):
                                            try:
                                                os.makedirs(self.target)
                                            except WindowsError, e:
                                                errors['destination']='directory not found: %s (%s)' % (self.target, WindowsErrorDecode(e))
                                    finally:
                                        win32wnet.WNetCancelConnection2(param_encode(unc, manager.default_encoding), 0, False)
                                except Exception, e:
                                    errors['destination']='directory not found: %s (%s)' % (self.target, WindowsErrorDecode(e))
                            else:
                                errors['destination']='directory not found: %s (%s)' % (self.target, WindowsErrorDecode(e))

        if not errors:

            if self.type!=None and self.type!='none':
                
                self.wba_args=[ self.exe, 'start', 'backup', '-quiet' ]
                # TODO: remove this
                
                if self.target:
                    self.wba_args.append('-backupTarget:%s' % ( self.target, ))
                    
                if self.include:
                    self.wba_args.append('-include:%s' % ( self.include, ))
                
                if self.target.startswith('\\\\') and self.user and self.password:
                    self.wba_args+=[ '-user:%s' % ( self.user, ), '-password:%s' % ( self.password,) ]
                    
                if self.type.lower()=='vssfull':
                    self.wba_args.append('-vssFull')
                elif self.type.lower()=='vsscopy':
                    self.wba_args.append('-vssCopy')
                    
                self.wba_args.extend(self.options)
                extra+='cmdline=%s\n' % self.hide_password('  '.join(self.wba_args))
    
        return errors, warnings, extra

    # -----------------------------------------------------------------------
    def run(self, command, job, manager):

        log=manager.log
        now=manager.now

        import windows

        start=int(time.time())
        wba_del_process=None           
        if self.name=='wbadminsys' and self.keepversions:
            wba_del_args=[ self.exe, 'delete', 'systemstatebackup', '-quiet', '-keepVersions:%d'%(self.keepversions-1,), '-backupTarget:%s' % ( self.target, ), '-machine:%s' % ( self.machine, ),   ]
            log.info('run: %s', self.hide_password(' '.join(wba_del_args)))
            try:
                wba_del_process=subprocess.Popen(param_encode(wba_del_args, manager.default_encoding), executable=self.wbadmin_bin, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                wba_del_out, wba_del_err=wba_del_process.communicate()
            except Exception, e:
                log.error('Unexpected exception running command: %s', self.hide_password(' '.join(wba_del_args)))
                raise
            else:
                if wba_del_process.returncode!=0:
                    log.error('Error code=%d running command: %s', wba_del_process.returncode, self.hide_password(' '.join(wba_del_args)))
            time.sleep(2)
        
        start2=int(time.time())
        log.info('run: %s', self.hide_password(' '.join(self.wba_args)))
        try:
            wba_process=subprocess.Popen(param_encode(self.wba_args, manager.default_encoding), executable=self.wbadmin_bin, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            wba_out, wba_err=wba_process.communicate()
        except Exception, e:
            log.error('Unexpected exception running command: %s', self.hide_password(' '.join(self.wba_args)))
            raise
        else:
            if wba_process.returncode!=0:
                log.error('Error code=%d running command: %s', wba_process.returncode, self.hide_password(' '.join(self.wba_args)))

        end=int(time.time())
        
        # get the Events
        # Sometime I miss the last event, retry 3 times at least
        time.sleep(1)
        count=3 # try only once now
        
        ev_status, hresult, msg, backupfailurelog='ERR', '', '', None
        while count>0:
            count-=1 
            wevt_args=[ 'wevtutil.exe', 'qe', 'Microsoft-Windows-Backup', '/rd:true', '/e:root', '/f:RenderedXml',# '/f:XML',
                        '/q:*[System[TimeCreated[timediff(@SystemTime)<=%d]]]' % (1000*int(time.time()-start2+1), ), 
                      ]  
            log.info('run: %s', ' '.join(wevt_args))
            wevt_process=subprocess.Popen(param_encode(wevt_args, manager.default_encoding), executable=self.wevtutil_bin, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            wevt_out, wevt_err=wevt_process.communicate()

            if not wevt_out:
                log.info('wevtutil output is empty exit=%d ERR[:128]=%r', wevt_process.returncode, wevt_err[:128])
            else:
                dom=xml.dom.minidom.parseString(wevt_out.decode(manager.default_encoding).encode('utf-8'))
                # search all Event, for HRESULT in EventID==14
                computer=platform.node()
                # TODO: check if computer name match
                for node in dom.getElementsByTagName('Event'):
                    system=node.getElementsByTagName('System')[0]
                    if getvalue(system, 'EventID')=='14':
                        eventdata=node.getElementsByTagName('EventData')[0]
                        for data in eventdata.getElementsByTagName('Data'):
                            if data.getAttribute('Name')=='HRESULT':
                                hresult=data.childNodes[0].data
                            elif data.getAttribute('Name')=='BackupFailureLogPath':
                                try:
                                    backupfailurelog=data.childNodes[0].data
                                except IndexError:
                                    pass
                                
                        renderinginfo=node.getElementsByTagName('RenderingInfo')[0]
                        msg=getvalue(renderinginfo, 'Message')
                    break
                
                if hresult:
                    break

            if count>0:
                log.info('EventID 14 missed, retry')
                time.sleep(21)
            else:
                if wba_process.returncode==0:
                    log.warning('EventID 14 not found')
                else:
                    log.info('EventID 14 not found')
                    
        if wba_process.returncode==0 and hresult=='0':
            ev_status='OK'
            flog=log.debug
        else:
            flog=log.info

        for line in wba_out.decode(manager.console_encoding).split('\n'):
            flog('> %s', line)

        status=u''
        status+='name=%s\r\n' % job['name']
        status+='program=%s\r\n' % self.name
        status+='version=%s\r\n' % self.__version__
        status+='hostname=%s\r\n' % platform.node()

        status+='status=%s\r\n' % ev_status
        
        status+='exit_code=%r\r\n' % wba_process.returncode
        status+='hresult=%s\r\n' % hresult
        status+='message=%s\r\n' % msg
        
        total_free_bytes, dir_out, dir_up_out, total_size=None, None, None, None
        
        if self.target:
            status+='target=%s\r\n' % self.target
            
            if self.target[-1]==':':
                target_dir=os.path.join(self.target+'\\', 'WindowsImageBackup', platform.node())
            else:
                target_dir=os.path.join(self.target, 'WindowsImageBackup', platform.node())
            if self.name=='wbadminsys':
                target_dir=os.path.join(target_dir, 'SystemStateBackup')
                
            # print 'ASX target_dir', target_dir
    
            # free_space
            try:
                target_full_filename=''
                total_free_bytes=free_space(self.target, log)
                # Search the Backup directory 
                dir_up_out, _ts=windows_list_dir(os.path.dirname(target_dir), log, manager.default_encoding)
                for filename in os.listdir(target_dir): #
                    full_filename=os.path.join(target_dir, filename)
                    stat=os.stat(full_filename)
                    if os.path.isdir(full_filename) and filename.lower().startswith('backup '):
                        if target_full_filename<full_filename:
                            target_full_filename=full_filename

                if target_full_filename:
                    dir_out, total_size=windows_list_dir(target_full_filename, log, manager.default_encoding)  
               
            except Exception, e:
                unc=os.path.splitunc(self.target)[0]
                if not unc or not self.user or not self.password:
                    log.error('checking target directory: %s', WindowsErrorDecode(e, manager.default_encoding))
                    status+='target_free_space=%s\r\n' % WindowsErrorDecode(e, manager.default_encoding)
                else:
                    try:
                        win32wnet.WNetAddConnection2(win32netcon.RESOURCETYPE_DISK, None, param_encode(self.target, manager.default_encoding), None, param_encode(self.user, manager.default_encoding), param_encode(self.password, manager.default_encoding))
                        try:
                            total_free_bytes=free_space(self.target, log)
                            dir_up_out, _ts=windows_list_dir(os.path.dirname(target_dir), log, manager.default_encoding)
                            for filename in os.listdir(target_dir): #
                                full_filename=os.path.join(target_dir, filename)
                                stat=os.stat(full_filename)
                                if os.path.isdir(full_filename) and filename.lower().startswith('backup '):
                                    dir_out, total_size=windows_list_dir(full_filename, log, manager.default_encoding)
                                    break  
                        finally:
                            win32wnet.WNetCancelConnection2(param_encode(self.target, manager.default_encoding), 0, False)
                    except Exception, e:
                        log.error('checking target directory: %s', WindowsErrorDecode(e, manager.default_encoding))
                        status+='target_free_space=%s\r\n' % WindowsErrorDecode(e, manager.default_encoding)
    
            if total_free_bytes:
                status+='target_free_space=%d\r\n' % total_free_bytes
            if total_size:
                status+='target_size=%d\r\n' % total_size

        # run wbadmin get versions
        wba_versions_args=[ self.exe, 'get', 'versions', ]
        #if self.target:
        #    wba_versions_args.append('-backupTarget:%s' % ( self.target, ))
        #    wba_versions_args.append('-machine:%s' % (self.machine, ))
        log.info('run: %s', ' '.join(wba_versions_args))
        wba_versions_process=subprocess.Popen(param_encode(wba_versions_args, manager.default_encoding), executable=self.wbadmin_bin, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        wba_versions_out, wba_versions_err=wba_versions_process.communicate()

        status+='start=%s\r\n' % time.ctime(start)
        status+='end=%s\r\n' % time.ctime(end)
        status+='start_epoch=%d\r\n' % start
        status+='end_epoch=%d\r\n' % end
        status+='wba_cmd=%s\r\n' % self.hide_password(' '.join(self.wba_args))
        status+='wevt_cmd=%s\r\n' % ' '.join(wevt_args)
        status+='wevt_exit_code=%s\r\n' % wevt_process.returncode
        status+='wba_get_versions_cmd=%s\r\n' % ' '.join(wba_versions_args)
        status+='wba_get_versions_exit_code=%s\r\n' % wba_versions_process.returncode
        
        if wba_del_process:
            
            status+='wba_del_cmd=%s\r\n' % ' '.join(wba_del_args)
            status+='wba_del_exit_code=%s\r\n' % wba_del_process.returncode
        
        attachments=[ # the type, subtype and coding is not used
                    ]
        
        if dir_up_out:
            attachments.append(('dir_up.txt', None, dir_up_out.encode('utf-8'), 'text', 'plain', 'utf-8'))

        if dir_out:
            attachments.append(('dir.txt', None, dir_out.encode('utf-8'), 'text', 'plain', 'utf-8'))
            
        if wba_out:
            attachments.append(('wba_out.txt', None, wba_out, 'text', 'plain', manager.console_encoding))
        if wba_err:
            attachments.append(('wba_err.txt', None, wba_err, 'text', 'plain', manager.console_encoding))

        if wba_del_process and wba_del_out:
            attachments.append(('wba_del_out.txt', None, wba_del_out, 'text', 'plain', manager.console_encoding))
        if wba_del_process and wba_del_err:
            attachments.append(('wba_del_err.txt', None, wba_del_err, 'text', 'plain', manager.console_encoding))

        if wba_versions_out:
            attachments.append(('wba_versions_out.txt', None, wba_versions_out, 'text', 'plain', manager.console_encoding))
        if wba_versions_err:
            attachments.append(('wba_versions_err.txt', None, wba_versions_err, 'text', 'plain', manager.console_encoding))
        
        if wevt_out:
            attachments.append(('wevt_out.txt', None, wevt_out, 'text', 'plain', manager.default_encoding))
        if wevt_err:
            attachments.append(('wevt_err.txt', None, wevt_err, 'text', 'plain', manager.default_encoding))
 
        if backupfailurelog:
            try:
                value=open(backupfailurelog, 'r').read()
            except Exception, e:
               log.error('reading backup failure log "%s": %s', backupfailurelog, WindowsErrorDecode(e, manager.default_encoding))
            else:
                if len(value)>2: # BOM
                    status+='backup_failure_log=%s\r\n' % backupfailurelog
                    attachments.append(('backupfailurelog.txt', None, value, 'text', 'plain', 'utf-16'))
               
        return ev_status, status, attachments


# ===========================================================================
#
# ===========================================================================
class WbadminSys(Wbadmin):

    name='wbadminsys'
    exe='wbadmin.exe'
    
    types=dict(systemstate='systemstate')

    # -----------------------------------------------------------------------
    def load(self, job, manager):
        
        log=manager.log
        now=manager.now

        errors, warnings, extra={}, {}, ''

        #
        # check job config
        #

        self.destination=job.get('destination', None)
        if not self.destination:
            errors['destination']='destination required'
        else:
            self.destination=self.destination.replace('\n','')
            try:
                self.destination=Destinations(self.destination, self)
            except (DestinationSyntaxError, cron.CronException), e:
                errors['destination']='syntax error: %s' % (str(e), )
            
        self.keepversions=job.get('keepversions', None)
        if self.keepversions:
            try:
                self.keepversions=int(self.keepversions)
                if self.keepversions<1:
                    errors['keepversions']='must be >= 1'
            except Exception:
                errors['keepversions']='must be an integer'

        if not errors:

            if self.destination:
                self.type, self.target=self.destination.match(now, self.night_shift)
    
            if self.type!=None and self.type!='none':
                
                self.wba_args=[ self.exe, 'start', 'systemstatebackup', '-quiet', '-backupTarget:%s' % ( self.target, ) ] 
                
                extra+='cmdline=%s\n' % '  '.join(self.wba_args)
    
        return errors, warnings, extra


# http://social.technet.microsoft.com/Forums/en/windowsbackup/thread/1437d4ab-e0eb-40af-953c-ca2d0b5d951f
#The backup versions are stored on the VSS shadow copies on the backup volume. 
#If space for the current backup is not enough, the older backup shadow copy 
#is reclaimed to make space for the new backup. It is recommended that the 
#backup target free space be at least 1.5 times the backup sources volume's 
#space, and the backup target be used for only storing backups 
#(other data churn on the backup target, may cause loss of shadowcopies, hence 
#loss of backup versions). Increasing the space allocated for shadow copies can 
#increase the number of backups avaliable. This can be done using the 
#vssadmin.exe utility, the command below will enable all the space in the 
#backup target to be used for shadow copies.
#
#vssadmin resize shadowstorage /for=d: /on=d: /maxsize=unbounded 
#
#Note: If Faster backup performance setting is turned on, Windows will leave
#a shadow copy on the source volume to use to track the changes. During next 
#backup operation, only the changes since last the backup are transferred
#(by reading from the "diff area" of the source volume shadow copy)-as 
#compared to the Normal backup performance option, where the entire source 
#volume data is transferred. (Because only changed blocks are written, there 
#is a performance increase.) The space used on the backup storage location 
#will still only be for the changed blocks detected on the source. Use this 
#method for servers that are less I/O intensive because shadow copies can 
#degrade the performance of write operations for the volume they are on 
#(read operations are not affected).
#

# http://social.technet.microsoft.com/Forums/en/windowsbackup/thread/3d4b3a2a-3f4f-4866-8d05-9baa86279e18
#You use the VSS Full Backup with VSS Aware applications like SQL Server, 
#Exchange Server to clear the transaction logs when the backup completed 
#succesfully. This is what they mean with 'Clearing application logs'.
#
#The best way to explain this is: Other backup applications clear the Archive 
#bit when the file is backed up, so they know this file has already been done. 
#When the file gets changed in any way after the backup, the archive bit is 
#set again so the backup application knows that the file needs to backed up.
#
#The VSS Copy Backup doesn't do anything with the archive bit, if it's on, 
#it's on. If it's off, it's off. If you use another product to backup the 
#same volume, use the VSS copy backup. If Windows Server Backup is the only 
#program you use, use the VSS Full Backup.

# http://social.technet.microsoft.com/Forums/en/windowsbackup/thread/d0fe1d25-c425-4491-af0e-e6ac199d89a1
#In Windows Server 2008 R2, there are several reasons which may cause the 
#windows backup to be full instead of incremental.
#
#1. Backup on the target is deleted/not present.
#2. Source volume VSS snapshot is deleted, from which the last backup was 
#taken, this can happen if the diff area on the source volume was reset, or 
#high data churn happen causing the older shadow opy to be claimed.
#3. 14 days have passed or 6 incremental backups have happened since the last 
#full backup.
#4. Churn on the backup source is high (more than 50%)
#
#Decision to do full vs incremental is a run-time decision that windows backups 
#takes. If all the conditions necessary for an incremental backup are not met, 
#a full backup is taken even if incremental backup settings are set.


# http://technet.microsoft.com/en-us/library/ee344835%28WS.10%29.aspx?ppud=4
# What's New in Windows Server Backup

# http://www.smaro.ch/blogs/Lists/Beitraege/Post.aspx?ID=772
# Backup Version and Space Management in Windows Server Backup

# http://www.tipandtrick.net/2008/how-to-suppress-and-bypass-system32-file-system-redirect-to-syswow64-folder-with-sysnative/
# How to Suppress and Bypass System32 File System Redirect to SysWOW64 Folder with Sysnative

# to install Windows Server Backup (command line)
# servermanagercmd.exe -install Backup-Features

# http://technet.microsoft.com/en-us/library/dd364735(WS.10).aspx
# There can be two ways to check if the backup had failed from the events:
#1. For every backup run, successful or failed, event-id 14 is generated. 
# The event-data for this ID contains the overall status of the backup in the 
# field 'HRESULT'. This should be 0 for a successful backup, non-zero otherwise.
#


# 
# http://social.technet.microsoft.com/Forums/en-US/windowsbackup/thread/13bb79e9-c970-4805-88cb-2000b3348070
# when backupping 2008 R2 on a samba server, share must have option 
# "strict allocate = yes" set
# Then remove existing file, they are "sparce/corrupted" and will make next backup fail 




# moflextab.py. This file automatically created by PLY (version 2.3). Don't edit!
_lextokens    = {'DT_UINT8': None, 'DISABLEOVERRIDE': None, 'PRAGMA': None, 'DT_SINT64': None, 'octalValue': None, 'hexValue': None, 'stringValue': None, 'decimalValue': None, 'NULL': None, 'PARAMETER': None, 'charValue': None, 'DT_DATETIME': None, 'REFERENCE': None, 'RESTRICTED': None, 'TOSUBCLASS': None, 'DT_REAL64': None, 'TRANSLATABLE': None, 'ENABLEOVERRIDE': None, 'DT_UINT32': None, 'ASSOCIATION': None, 'SCHEMA': None, 'DT_BOOL': None, 'floatValue': None, 'METHOD': None, 'DT_SINT16': None, 'AS': None, 'DT_SINT32': None, 'binaryValue': None, 'DT_STR': None, 'SCOPE': None, 'FLAVOR': None, 'DT_CHAR16': None, 'TRUE': None, 'DT_REAL32': None, 'DT_SINT8': None, 'FALSE': None, 'QUALIFIER': None, 'DT_UINT64': None, 'OF': None, 'REF': None, 'CLASS': None, 'INSTANCE': None, 'IDENTIFIER': None, 'PROPERTY': None, 'INDICATION': None, 'DT_UINT16': None, 'ANY': None}
_lexreflags   = 0
_lexliterals  = '#(){};[],$:='
_lexstateinfo = {'INITIAL': 'inclusive'}
_lexstatere   = {'INITIAL': [('(?P<t_COMMENT>//.*)|(?P<t_MCOMMENT>/\\*(.|\\n)*?\\*/)|(?P<t_IDENTIFIER>([a-zA-Z_]|(([\\xC2-\\xDF][\\x80-\\xBF])|(\\xE0[\\xA0-\\xBF][\\x80-\\xBF])|([\\xE1-\\xEC][\\x80-\\xBF][\\x80-\\xBF])|(\\xED[\\x80-\\x9F][\\x80-\\xBF])|([\\xEE-\\xEF][\\x80-\\xBF][\\x80-\\xBF])|(\\xF0[\\x90-\\xBF][\\x80-\\xBF][\\x80-\\xBF])|([\\xF1-\\xF3][\\x80-\\xBF][\\x80-\\xBF][\\x80-\\xBF])|(\\xF4[\\x80-\\x8F][\\x80-\\xBF][\\x80-\\xBF])))([0-9a-zA-Z_]|(([\\xC2-\\xDF][\\x80-\\xBF])|(\\xE0[\\xA0-\\xBF][\\x80-\\xBF])|([\\xE1-\\xEC][\\x80-\\xBF][\\x80-\\xBF])|(\\xED[\\x80-\\x9F][\\x80-\\xBF])|([\\xEE-\\xEF][\\x80-\\xBF][\\x80-\\xBF])|(\\xF0[\\x90-\\xBF][\\x80-\\xBF][\\x80-\\xBF])|([\\xF1-\\xF3][\\x80-\\xBF][\\x80-\\xBF][\\x80-\\xBF])|(\\xF4[\\x80-\\x8F][\\x80-\\xBF][\\x80-\\xBF])))*)|(?P<t_newline>\\n+)|(?P<t_stringValue>"([^"\\\\\\n\\r]|([\\\\](([bfnrt\'"\\\\])|(x[0-9a-fA-F]{1,4}))))*")|(?P<t_floatValue>[+-]?[0-9]*\\.[0-9]+([eE][+-]?[0-9]+)?)|(?P<t_hexValue>[+-]?0[xX][0-9a-fA-F]+)|(?P<t_decimalValue>[+-]?([1-9][0-9]*|0))|(?P<t_binaryValue>[+-]?[01]+[bB])|(?P<t_octalValue>[+-]?0[0-7]+)', [None, ('t_COMMENT', 'COMMENT'), ('t_MCOMMENT', 'MCOMMENT'), None, ('t_IDENTIFIER', 'IDENTIFIER'), None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, ('t_newline', 'newline'), (None, 'stringValue'), None, None, None, None, None, (None, 'floatValue'), None, (None, 'hexValue'), (None, 'decimalValue'), None, (None, 'binaryValue'), (None, 'octalValue')])]}
_lexstateignore = {'INITIAL': ' \r\t'}
_lexstateerrorf = {'INITIAL': 't_error'}



#!/usr/bin/env python
#
# (C) Copyright 2006-2007 Novell, Inc. 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2 of the
# License.
#   
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#   
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Author: Bart Whiteley <bwhiteley suse.de>

import sys
import os
import lex
import yacc
from lex import TOKEN
from cim_operations import CIMError, WBEMConnection
from cim_obj import *
from cim_constants import *
from getpass import getpass

_optimize = 1
_tabmodule='mofparsetab'
_lextab='moflextab'


reserved = {
    'any':'ANY',
    'as':'AS',
    'association':'ASSOCIATION',
    'class':'CLASS',
    'disableoverride':'DISABLEOVERRIDE',
    'boolean':'DT_BOOL',
    'char16':'DT_CHAR16',
    'datetime':'DT_DATETIME',
    'pragma':'PRAGMA',
    'real32':'DT_REAL32',
    'real64':'DT_REAL64',
    'sint16':'DT_SINT16',
    'sint32':'DT_SINT32',
    'sint64':'DT_SINT64',
    'sint8':'DT_SINT8',
    'string':'DT_STR',
    'uint16':'DT_UINT16',
    'uint32':'DT_UINT32',
    'uint64':'DT_UINT64',
    'uint8':'DT_UINT8',
    'enableoverride':'ENABLEOVERRIDE',
    'false':'FALSE',
    'flavor':'FLAVOR',
    'indication':'INDICATION',
    'instance':'INSTANCE',
    'method':'METHOD',
    'null':'NULL',
    'of':'OF',
    'parameter':'PARAMETER',
    'property':'PROPERTY',
    'qualifier':'QUALIFIER',
    'ref':'REF',
    'reference':'REFERENCE',
    'restricted':'RESTRICTED',
    'schema':'SCHEMA',
    'scope':'SCOPE',
    'tosubclass':'TOSUBCLASS',
    'translatable':'TRANSLATABLE',
    'true':'TRUE',
    }

tokens = reserved.values() + [
        'IDENTIFIER',
        'stringValue',
        'floatValue',
        'charValue',
        'binaryValue',
        'octalValue',
        'decimalValue',
        'hexValue',
    ]

literals = '#(){};[],$:='

# UTF-8 (from Unicode 4.0.0 standard):
# Table 3-6. Well-Formed UTF-8 Byte Sequences Code Points
# 1st Byte 2nd Byte 3rd Byte 4th Byte
# U+0000..U+007F     00..7F
# U+0080..U+07FF     C2..DF   80..BF
# U+0800..U+0FFF     E0       A0..BF   80..BF
# U+1000..U+CFFF     E1..EC   80..BF   80..BF
# U+D000..U+D7FF     ED       80..9F   80..BF
# U+E000..U+FFFF     EE..EF   80..BF   80..BF
# U+10000..U+3FFFF   F0       90..BF   80..BF   80..BF
# U+40000..U+FFFFF   F1..F3   80..BF   80..BF   80..BF
# U+100000..U+10FFFF F4       80..8F   80..BF   80..BF

utf8_2 = r'[\xC2-\xDF][\x80-\xBF]'
utf8_3_1 = r'\xE0[\xA0-\xBF][\x80-\xBF]'
utf8_3_2 = r'[\xE1-\xEC][\x80-\xBF][\x80-\xBF]'
utf8_3_3 = r'\xED[\x80-\x9F][\x80-\xBF]'
utf8_3_4 = r'[\xEE-\xEF][\x80-\xBF][\x80-\xBF]'
utf8_4_1 = r'\xF0[\x90-\xBF][\x80-\xBF][\x80-\xBF]'
utf8_4_2 = r'[\xF1-\xF3][\x80-\xBF][\x80-\xBF][\x80-\xBF]'
utf8_4_3 = r'\xF4[\x80-\x8F][\x80-\xBF][\x80-\xBF]'

utf8Char = r'(%s)|(%s)|(%s)|(%s)|(%s)|(%s)|(%s)|(%s)' % (utf8_2, utf8_3_1,
        utf8_3_2, utf8_3_3, utf8_3_4, utf8_4_1, utf8_4_2, utf8_4_3)

def t_COMMENT(t):
    r'//.*'
    pass

def t_MCOMMENT(t):
    r'/\*(.|\n)*?\*/'
    t.lineno += t.value.count('\n')


t_binaryValue = r'[+-]?[01]+[bB]'
t_octalValue = r'[+-]?0[0-7]+'
t_decimalValue = r'[+-]?([1-9][0-9]*|0)'
t_hexValue = r'[+-]?0[xX][0-9a-fA-F]+'
t_floatValue = r'[+-]?[0-9]*\.[0-9]+([eE][+-]?[0-9]+)?'

simpleEscape = r"""[bfnrt'"\\]"""
hexEscape = r'x[0-9a-fA-F]{1,4}'
escapeSequence = r'[\\]((%s)|(%s))' % (simpleEscape, hexEscape)
cChar = r"[^'\\\n\r]|(%s)" % escapeSequence
sChar = r'[^"\\\n\r]|(%s)' % escapeSequence
charValue = r"'%s'" % cChar

t_stringValue = r'"(%s)*"' % sChar

identifier_re = r'([a-zA-Z_]|(%s))([0-9a-zA-Z_]|(%s))*' % (utf8Char, utf8Char)

@TOKEN(identifier_re)
def t_IDENTIFIER(t):
    t.type = reserved.get(t.value.lower(),'IDENTIFIER') # check for reserved word
    return t

# Define a rule so we can track line numbers
def t_newline(t):
    r'\n+'
    t.lexer.lineno += len(t.value)
    t.lexer.linestart = t.lexpos

t_ignore = ' \r\t'

# Error handling rule
def t_error(t):
    msg = "Illegal character '%s' " % t.value[0]
    msg+= "Line %d, col %d" % (t.lineno, _find_column(t.lexer.parser.mof, t))
    t.lexer.parser.log(msg)
    t.lexer.skip(1)

class MOFParseError(ValueError):
    pass

def p_error(p):
    ex = MOFParseError()
    if p is None:
        ex.args = ('Unexpected end of file',)
        raise ex
    ex.file = p.lexer.parser.file
    ex.lineno = p.lineno
    ex.column = _find_column(p.lexer.parser.mof, p)
    ex.context = _get_error_context(p.lexer.parser.mof, p)
    raise ex


def p_mofSpecification(p):
    """mofSpecification : mofProductionList"""

def p_mofProductionList(p):
    """mofProductionList : empty
                         | mofProductionList mofProduction
                           """

def p_mofProduction(p):
    """mofProduction : compilerDirective
                     | mp_createClass
                     | mp_setQualifier
                     | mp_createInstance
                     """


def _create_ns(p, handle, ns):
    # Figure out the flavor of cim server
    cimom_type = None
    ns = ns.strip('/')
    try:
        inames = handle.EnumerateInstanceNames('__Namespace', namespace='root')
        inames = [x['name'] for x in inames]
        if 'PG_InterOp' in inames:
            cimom_type = 'pegasus'
    except CIMError, ce:
        if ce.args[0] != CIM_ERR_NOT_FOUND:
            ce.file_line = (p.parser.file, p.lexer.lineno)
            raise
    if not cimom_type:
        try:
            inames = handle.EnumerateInstanceNames('CIM_Namespace', 
                    namespace='Interop')
            inames = [x['name'] for x in inames]
            cimom_type = 'proper'
        except CIMError, ce:
            ce.file_line = (p.parser.file, p.lexer.lineno)
            raise

    if not cimom_type:
        ce = CIMError(CIM_ERR_FAILED, 
                'Unable to determine CIMOM type')
        ce.file_line = (p.parser.file, p.lexer.lineno)
        raise ce
    if cimom_type == 'pegasus':
        # To create a namespace in Pegasus, create an instance of 
        # __Namespace with  __Namespace.Name = '', and create it in 
        # the target namespace to be created. 
        inst = CIMInstance('__Namespace', 
                properties={'Name':''},
                path=CIMInstanceName('__Namespace', 
                    keybindings={'Name':''},
                    namespace=ns))
        try:
            handle.CreateInstance(inst)
        except CIMError, ce:
            if ce.args[0] != CIM_ERR_ALREADY_EXISTS:
                ce.file_line = (p.parser.file, p.lexer.lineno)
                raise

    elif cimom_type == 'proper':
        inst = CIMInstance('CIM_Namespace', 
                properties={'Name': ns},
                path=CIMInstanceName('CIM_Namespace',
                    namespace='root',
                    keybindings={'Name':ns}))
        handle.CreateInstance(inst)




def p_mp_createClass(p):
    """mp_createClass : classDeclaration
                      | assocDeclaration
                      | indicDeclaration
                      """
    ns = p.parser.handle.default_namespace
    cc = p[1]
    try:
        fixedNS = fixedRefs = fixedSuper = False
        while not fixedNS or not fixedRefs or not fixedSuper:
            try:
                if p.parser.verbose:
                    p.parser.log('Creating class %s:%s' % (ns, cc.classname))
                p.parser.handle.CreateClass(cc)
                if p.parser.verbose:
                    p.parser.log('Created class %s:%s' % (ns,cc.classname))
                p.parser.classnames[ns].append(cc.classname.lower())
                break
            except CIMError, ce:
                ce.file_line = (p.parser.file, p.lexer.lineno)
                errcode = ce.args[0]
                if errcode == CIM_ERR_INVALID_NAMESPACE:
                    if fixedNS:
                        raise
                    if p.parser.verbose:
                        p.parser.log('Creating namespace ' + ns)
                    _create_ns(p, p.parser.handle, ns)
                    fixedNS = True
                    continue
                if not p.parser.search_paths:
                    raise
                if errcode == CIM_ERR_INVALID_SUPERCLASS:
                    if fixedSuper:
                        raise
                    moffile = p.parser.mofcomp.find_mof(cc.superclass)
                    if not moffile:
                        raise
                    p.parser.mofcomp.compile_file(moffile, ns)
                    fixedSuper = True
                elif errcode in [CIM_ERR_INVALID_PARAMETER,
                                 CIM_ERR_NOT_FOUND,
                                 CIM_ERR_FAILED]:
                    if fixedRefs:
                        raise
                    if not p.parser.qualcache[ns]:
                        for fname in ['qualifiers', 'qualifiers_optional']:
                            qualfile = p.parser.mofcomp.find_mof(fname)
                            if qualfile:
                                p.parser.mofcomp.compile_file(qualfile,ns)
                    if not p.parser.qualcache[ns]:
                        # can't find qualifiers
                        raise
                    objects = cc.properties.values()
                    for meth in cc.methods.values():
                        objects+= meth.parameters.values()
                    dep_classes = []
                    for obj in objects:
                        if obj.type not in ['reference','string']:
                            continue
                        if obj.type == 'reference':
                            if obj.reference_class.lower() not in dep_classes:
                                dep_classes.append(obj.reference_class.lower())
                            continue
                        # else obj.type is 'string'
                        try:
                            embedded_inst= obj.qualifiers['embeddedinstance']
                        except KeyError:
                            continue
                        embedded_inst = embedded_inst.value.lower()
                        if embedded_inst not in dep_classes:
                            dep_classes.append(embedded_inst)
                        continue
                    for klass in dep_classes:
                        if klass in p.parser.classnames[ns]:
                            continue
                        try:
                            # don't limit it with LocalOnly=True, 
                            # PropertyList, IncludeQualifiers=False, ...
                            # because of caching in case we're using the
                            # special WBEMConnection subclass used for 
                            # removing schema elements
                            p.parser.handle.GetClass(klass,
                                    LocalOnly=False, 
                                    IncludeQualifiers=True)
                            p.parser.classnames[ns].append(klass)
                        except CIMError:
                            moffile = p.parser.mofcomp.find_mof(klass)
                            if not moffile:
                                raise
                            p.parser.mofcomp.compile_file(moffile, ns)
                            p.parser.classnames[ns].append(klass)
                    fixedRefs = True
                else:
                    raise
    
    except CIMError, ce:
        ce.file_line = (p.parser.file, p.lexer.lineno)
        if ce.args[0] != CIM_ERR_ALREADY_EXISTS:
            raise
        if p.parser.verbose:
            p.parser.log('Class %s already exist.  Modifying...' % cc.classname)
        try:
            p.parser.handle.ModifyClass(cc, ns)
        except CIMError, ce:
            p.parser.log('Error Modifying class %s: %s, %s' 
                    % (cc.classname, ce.args[0], ce.args[1]))

def p_mp_createInstance(p):
    """mp_createInstance : instanceDeclaration"""
    inst = p[1]
    if p.parser.verbose:
        p.parser.log('Creating instance of %s.' % inst.classname)
    try:
        p.parser.handle.CreateInstance(inst)
    except CIMError, ce:
        if ce.args[0] == CIM_ERR_ALREADY_EXISTS:
            if p.parser.verbose:
                p.parser.log('Instance of class %s already exist.  Modifying...'  % inst.classname)
            try:
                p.parser.handle.ModifyInstance(inst)
            except CIMError, ce:
                if ce.args[0] == CIM_ERR_NOT_SUPPORTED:
                    if p.parser.verbose:
                        p.parser.log('ModifyInstance not supported.  Deleting instance of %s: %s' % (inst.classname, inst.path))
                    p.parser.handle.DeleteInstance(inst.path)
                    if p.parser.verbose:
                        p.parser.log('Creating instance of %s.' % inst.classname)
                    p.parser.handle.CreateInstance(inst)
        else:
            ce.file_line = (p.parser.file, p.lexer.lineno)
            raise

def p_mp_setQualifier(p):
    """mp_setQualifier : qualifierDeclaration"""
    qualdecl = p[1]
    ns = p.parser.handle.default_namespace
    if p.parser.verbose:
        p.parser.log('Setting qualifier %s' % qualdecl.name)
    try:
        p.parser.handle.SetQualifier(qualdecl)
    except CIMError, ce:
        if ce.args[0] == CIM_ERR_INVALID_NAMESPACE:
            if p.parser.verbose:
                p.parser.log('Creating namespace ' + ns)
            _create_ns(p, p.parser.handle, ns)
            if p.parser.verbose:
                p.parser.log('Setting qualifier %s' % qualdecl.name)
            p.parser.handle.SetQualifier(qualdecl)
        elif ce.args[0] == CIM_ERR_NOT_SUPPORTED:
            if p.parser.verbose:
                p.parser.log('Qualifier %s already exists.  Deleting...' 
                            % qualdecl.name)
            p.parser.handle.DeleteQualifier(qualdecl.name)
            if p.parser.verbose:
                p.parser.log('Setting qualifier %s' % qualdecl.name)
            p.parser.handle.SetQualifier(qualdecl)
        else:
            ce.file_line = (p.parser.file, p.lexer.lineno)
            raise
    p.parser.qualcache[ns][qualdecl.name] = qualdecl

def p_compilerDirective(p): 
    """compilerDirective : '#' PRAGMA pragmaName '(' pragmaParameter ')'"""
    directive = p[3].lower()
    param = p[5]
    if directive == 'include':
        fname = param
        #if p.parser.file:
        fname = os.path.dirname(p.parser.file) + '/' + fname
        p.parser.mofcomp.compile_file(fname, p.parser.handle.default_namespace)
    elif directive == 'namespace':
        p.parser.handle.default_namespace = param
        if param not in p.parser.qualcache:
            p.parser.qualcache[param] = NocaseDict()
    
    p[0] = None

def p_pragmaName(p):
    """pragmaName : identifier"""
    p[0] = p[1]

def p_pragmaParameter(p):
    """pragmaParameter : stringValue"""
    p[0] = _fixStringValue(p[1])

def p_classDeclaration(p):
    """classDeclaration : CLASS className '{' classFeatureList '}' ';'
                        | CLASS className superClass '{' classFeatureList '}' ';'
                        | CLASS className alias '{' classFeatureList '}' ';'
                        | CLASS className alias superClass '{' classFeatureList '}' ';'
                        | qualifierList CLASS className '{' classFeatureList '}' ';'
                        | qualifierList CLASS className superClass '{' classFeatureList '}' ';'
                        | qualifierList CLASS className alias '{' classFeatureList '}' ';'
                        | qualifierList CLASS className alias superClass '{' classFeatureList '}' ';'
                        """
    superclass = None
    alias = None
    quals = []
    if isinstance(p[1], basestring): # no class qualifiers
        cname = p[2]
        if p[3][0] == '$': # alias present
            alias = p[3]
            if p[4] == '{': # no superclass
                cfl = p[5]
            else: # superclass
                superclass = p[4]
                cfl = p[6]
        else: # no alias
            if p[3] == '{': # no superclass
                cfl = p[4]
            else: # superclass
                superclass = p[3]
                cfl = p[5]
    else: # class qualifiers
        quals = p[1]
        cname = p[3]
        if p[4][0] == '$': # alias present
            alias = p[4]
            if p[5] == '{': # no superclass
                cfl = p[6]
            else: # superclass
                superclass = p[5]
                cfl = p[7]
        else: # no alias
            if p[4] == '{': # no superclass
                cfl = p[5]
            else: # superclass
                superclass = p[4]
                cfl = p[6]
    quals = dict([(x.name, x) for x in quals])
    methods = {}
    props = {}
    for item in cfl:
        item.class_origin = cname
        if isinstance(item, CIMMethod):
            methods[item.name] = item
        else:
            props[item.name] = item
    p[0] = CIMClass(cname, properties=props, methods=methods, 
            superclass=superclass, qualifiers=quals)
    if alias:
        p.parser.aliases[alias] = p[0]

def p_classFeatureList(p):
    """classFeatureList : empty
                        | classFeatureList classFeature
                        """
    if len(p) == 2:
        p[0] = []
    else:
        p[0] = p[1] + [p[2]]

def p_assocDeclaration(p):
    """assocDeclaration : '[' ASSOCIATION qualifierListEmpty ']' CLASS className '{' associationFeatureList '}' ';'
                        | '[' ASSOCIATION qualifierListEmpty ']' CLASS className superClass '{' associationFeatureList '}' ';'
                        | '[' ASSOCIATION qualifierListEmpty ']' CLASS className alias '{' associationFeatureList '}' ';'
                        | '[' ASSOCIATION qualifierListEmpty ']' CLASS className alias superClass '{' associationFeatureList '}' ';'
                        """
    aqual = CIMQualifier('ASSOCIATION', True, type='boolean')
    # TODO flavor trash. 
    quals = [aqual] + p[3]
    p[0] = _assoc_or_indic_decl(quals, p)
    
def p_indicDeclaration(p):
    """indicDeclaration : '[' INDICATION qualifierListEmpty ']' CLASS className '{' classFeatureList '}' ';'
                        | '[' INDICATION qualifierListEmpty ']' CLASS className superClass '{' classFeatureList '}' ';'
                        | '[' INDICATION qualifierListEmpty ']' CLASS className alias '{' classFeatureList '}' ';'
                        | '[' INDICATION qualifierListEmpty ']' CLASS className alias superClass '{' classFeatureList '}' ';'
                        """
    iqual = CIMQualifier('INDICATION', True, type='boolean')
    # TODO flavor trash. 
    quals = [iqual] + p[3]
    p[0] = _assoc_or_indic_decl(quals, p)

def _assoc_or_indic_decl(quals, p):
    """(refer to grammer rules on p_assocDeclaration and p_indicDeclaration)"""
    superclass = None
    alias = None
    cname = p[6]
    if p[7] == '{':
        cfl = p[8]
    elif p[7][0] == '$': # alias
        alias = p[7]
        if p[8] == '{':
            cfl = p[9]
        else:
            superclass = p[8]
            cfl = p[10]
    else:
        superclass = p[7]
        cfl = p[9]
    props = {}
    methods = {}
    for item in cfl:
        item.class_origin = cname
        if isinstance(item, CIMMethod):
            methods[item.name] = item
        else:
            props[item.name] = item
    quals = dict([(x.name, x) for x in quals])
    cc = CIMClass(cname, properties=props, methods=methods, 
            superclass=superclass, qualifiers=quals)
    if alias:
        p.parser.aliases[alias] = cc
    return cc

def p_qualifierListEmpty(p):
    """qualifierListEmpty : empty
                          | qualifierListEmpty ',' qualifier
                          """
    if len(p) == 2:
        p[0] = []
    else:
        p[0] = p[1] + [p[3]]

def p_associationFeatureList(p):
    """associationFeatureList : empty 
                              | associationFeatureList associationFeature
                              """
    if len(p) == 2:
        p[0] = []
    else:
        p[0] = p[1] + [p[2]]

def p_className(p):
    """className : identifier"""
    p[0] = p[1]

def p_alias(p):
    """alias : AS aliasIdentifier"""
    p[0] = p[2]

def p_aliasIdentifier(p):
    """aliasIdentifier : '$' identifier"""
    p[0] = '$%s' % p[2]

def p_superClass(p):
    """superClass : ':' className"""
    p[0] = p[2]

def p_classFeature(p):
    """classFeature : propertyDeclaration
                    | methodDeclaration
                    | referenceDeclaration
                    """
    p[0] = p[1]

def p_associationFeature(p):
    """associationFeature : classFeature"""
    p[0] = p[1]

def p_qualifierList(p):
    """qualifierList : '[' qualifier qualifierListEmpty ']'"""
    p[0] = [p[2]] + p[3]

def p_qualifier(p): 
    """qualifier : qualifierName
                 | qualifierName ':' flavorList
                 | qualifierName qualifierParameter
                 | qualifierName qualifierParameter ':' flavorList
                 """
    qname = p[1]
    ns = p.parser.handle.default_namespace
    qval = None
    flavorlist = []
    if len(p) == 3:
        qval = p[2]
    elif len(p) == 4:
        flavorlist = p[3]
    elif len(p) == 5:
        qval = p[2]
        flavorlist = p[4]
    try:
        qualdecl = p.parser.qualcache[ns][qname]
    except KeyError:
        try:
            quals = p.parser.handle.EnumerateQualifiers()
        except CIMError, ce:
            if ce.args[0] != CIM_ERR_INVALID_NAMESPACE:
                ce.file_line = (p.parser.file, p.lexer.lineno)
                raise
            _create_ns(p, p.parser.handle, ns)
            quals = None

        if quals:
            for qual in quals:
                p.parser.qualcache[ns][qual.name] = qual
        else:
            for fname in ['qualifiers', 'qualifiers_optional']:
                qualfile = p.parser.mofcomp.find_mof(fname)
                if qualfile:
                    p.parser.mofcomp.compile_file(qualfile, ns)
    try:
        qualdecl = p.parser.qualcache[ns][qname]
    except KeyError:
        ce = CIMError(CIM_ERR_FAILED, 'Unknown Qualifier: %s' % qname)
        ce.file_line = (p.parser.file, p.lexer.lineno)
        raise ce
        
    flavors = _build_flavors(flavorlist, qualdecl)
    if qval is None: 
        if qualdecl.type == 'boolean':
            qval = True
        else:
            qval = qualdecl.value # default value
    else:
        qval = tocimobj(qualdecl.type, qval)
    p[0] = CIMQualifier(qname, qval, type=qualdecl.type, **flavors)
    # TODO propagated? 

def p_flavorList(p):
    """flavorList : flavor
                  | flavorList flavor
                  """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[2]]

def p_qualifierParameter(p):
    """qualifierParameter : '(' constantValue ')'
                          | arrayInitializer
                          """
    if len(p) == 2:
        p[0] = p[1]
    else:
        p[0] = p[2]

def p_flavor(p):
    """flavor : ENABLEOVERRIDE
              | DISABLEOVERRIDE
              | RESTRICTED
              | TOSUBCLASS
              | TRANSLATABLE
              """
    p[0] = p[1].lower()

def p_propertyDeclaration(p):
    """propertyDeclaration : propertyDeclaration_1
                           | propertyDeclaration_2
                           | propertyDeclaration_3
                           | propertyDeclaration_4
                           | propertyDeclaration_5
                           | propertyDeclaration_6
                           | propertyDeclaration_7
                           | propertyDeclaration_8
                           """
    p[0] = p[1]

def p_propertyDeclaration_1(p):
    """propertyDeclaration_1 : dataType propertyName ';'"""
    p[0] = CIMProperty(p[2], None, type=p[1])

def p_propertyDeclaration_2(p):
    """propertyDeclaration_2 : dataType propertyName defaultValue ';'"""
    p[0] = CIMProperty(p[2], p[3], type=p[1])

def p_propertyDeclaration_3(p):
    """propertyDeclaration_3 : dataType propertyName array ';'"""
    p[0] = CIMProperty(p[2], None, type=p[1], is_array=True, 
            array_size=p[3])

def p_propertyDeclaration_4(p):
    """propertyDeclaration_4 : dataType propertyName array defaultValue ';'"""
    p[0] = CIMProperty(p[2], p[4], type=p[1], is_array=True, 
            array_size=p[3])

def p_propertyDeclaration_5(p):
    """propertyDeclaration_5 : qualifierList dataType propertyName ';'"""
    quals = dict([(x.name, x) for x in p[1]])
    p[0] = CIMProperty(p[3], None, type=p[2], qualifiers=quals)

def p_propertyDeclaration_6(p):
    """propertyDeclaration_6 : qualifierList dataType propertyName defaultValue ';'"""
    quals = dict([(x.name, x) for x in p[1]])
    p[0] = CIMProperty(p[3], tocimobj(p[2], p[4]), 
            type=p[2], qualifiers=quals)

def p_propertyDeclaration_7(p):
    """propertyDeclaration_7 : qualifierList dataType propertyName array ';'"""
    quals = dict([(x.name, x) for x in p[1]])
    p[0] = CIMProperty(p[3], None, type=p[2], qualifiers=quals,
            is_array=True, array_size=p[4])

def p_propertyDeclaration_8(p):
    """propertyDeclaration_8 : qualifierList dataType propertyName array defaultValue ';'"""
    quals = dict([(x.name, x) for x in p[1]])
    p[0] = CIMProperty(p[3], tocimobj(p[2], p[5]), 
            type=p[2], qualifiers=quals, is_array=True, array_size=p[4])

def p_referenceDeclaration(p):
    """referenceDeclaration : objectRef referenceName ';'
                            | objectRef referenceName defaultValue ';'
                            | qualifierList objectRef referenceName ';'
                            | qualifierList objectRef referenceName defaultValue ';'
                            """
    quals = []
    dv = None
    if isinstance(p[1], list): # qualifiers
        quals = p[1]
        cname = p[2]
        pname = p[3]
        if len(p) == 6:
            dv = p[4]
    else:
        cname = p[1]
        pname = p[2]
        if len(p) == 5:
            dv = p[3]
    quals = dict([(x.name, x) for x in quals])
    p[0] = CIMProperty(pname, dv, type='reference', 
            reference_class=cname, qualifiers=quals)

def p_methodDeclaration(p):
    """methodDeclaration : dataType methodName '(' ')' ';'
                         | dataType methodName '(' parameterList ')' ';'
                         | qualifierList dataType methodName '(' ')' ';'
                         | qualifierList dataType methodName '(' parameterList ')' ';'
                         """
    paramlist = []
    quals = []
    if isinstance(p[1], basestring): # no quals
        dt = p[1]
        mname = p[2]
        if p[4] != ')':
            paramlist = p[4]
    else: # quals present
        quals = p[1]
        dt = p[2]
        mname = p[3]
        if p[5] != ')':
            paramlist = p[5]
    params = dict([(param.name, param) for param in paramlist])
    quals = dict([(q.name, q) for q in quals])
    p[0] = CIMMethod(mname, return_type=dt, parameters=params, 
            qualifiers=quals)
    # note: class_origin is set when adding method to class. 
    # TODO what to do with propagated? 

def p_propertyName(p):
    """propertyName : identifier"""
    p[0] = p[1]

def p_referenceName(p):
    """referenceName : identifier"""
    p[0] = p[1]

def p_methodName(p):
    """methodName : identifier"""
    p[0] = p[1]

def p_dataType(p):
    """dataType : DT_UINT8
                | DT_SINT8
                | DT_UINT16
                | DT_SINT16
                | DT_UINT32
                | DT_SINT32
                | DT_UINT64
                | DT_SINT64
                | DT_REAL32
                | DT_REAL64
                | DT_CHAR16
                | DT_STR
                | DT_BOOL
                | DT_DATETIME
                """
    p[0] = p[1].lower()

def p_objectRef(p):
    """objectRef : className REF"""
    p[0] = p[1]

def p_parameterList(p):
    """parameterList : parameter
                     | parameterList ',' parameter
                     """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[3]]

def p_parameter(p):
    """parameter : parameter_1
                 | parameter_2
                 | parameter_3
                 | parameter_4
                 """
    p[0] = p[1]

def p_parameter_1(p):
    """parameter_1 : dataType parameterName
                   | dataType parameterName array
                   """
    args = {}
    if len(p) == 4:
        args['is_array'] = True
        args['array_size'] = p[3]
    p[0] = CIMParameter(p[2], p[1], **args)

def p_parameter_2(p):
    """parameter_2 : qualifierList dataType parameterName
                   | qualifierList dataType parameterName array
                   """
    args = {}
    if len(p) == 5:
        args['is_array'] = True
        args['array_size'] = p[4]
    quals = dict([(x.name, x) for x in p[1]])
    p[0] = CIMParameter(p[3], p[2], qualifiers=quals, **args)

def p_parameter_3(p):
    """parameter_3 : objectRef parameterName
                   | objectRef parameterName array
                   """
    args = {}
    if len(p) == 4:
        args['is_array'] = True
        args['array_size'] = p[3]
    p[0] = CIMParameter(p[2], 'reference', reference_class=p[1], **args)

def p_parameter_4(p):
    """parameter_4 : qualifierList objectRef parameterName
                   | qualifierList objectRef parameterName array
                   """
    args = {}
    if len(p) == 5:
        args['is_array'] = True
        args['array_size'] = p[4]
    quals = dict([(x.name, x) for x in p[1]])
    p[0] = CIMParameter(p[3], 'reference', qualifiers=quals, 
                reference_class=p[2], **args)

def p_parameterName(p):
    """parameterName : identifier"""
    p[0] = p[1]

def p_array(p):
    """array : '[' ']'
             | '[' integerValue ']'
             """
    if len(p) == 3:
        p[0] = None
    else:
        p[0] = p[2]

def p_defaultValue(p):
    """defaultValue : '=' initializer"""
    p[0] = p[2]

def p_initializer(p):
    """initializer : constantValue
                   | arrayInitializer
                   | referenceInitializer
                   """
    p[0] = p[1]

def p_arrayInitializer(p):
    """arrayInitializer : '{' constantValueList '}'
                        | '{' '}'
                        """
    if len(p) == 3:
        p[0] = []
    else:
        p[0] = p[2]

def p_constantValueList(p):
    """constantValueList : constantValue
                         | constantValueList ',' constantValue
                         """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[3]]

def _fixStringValue(s):
    s = s[1:-1]
    rv = ''
    esc = False
    i = -1 
    while i < len(s) -1:
        i+= 1
        ch = s[i]
        if ch == '\\' and not esc:
            esc = True
            continue
        if not esc:
            rv+= ch
            continue

        if ch == '"'   : rv+= '"'
        elif ch == 'n' : rv+= '\n'
        elif ch == 't' : rv+= '\t'
        elif ch == 'b' : rv+= '\b'
        elif ch == 'f' : rv+= '\f'
        elif ch == 'r' : rv+= '\r'
        elif ch == '\\': rv+= '\\'
        elif ch in ['x','X']:
            hexc = 0
            j = 0
            i+= 1
            while j < 4:
                c = s[i+j]; 
                c = c.upper()
                if not c.isdigit() and not c in 'ABCDEF':
                    break;
                hexc <<= 4
                if c.isdigit():
                    hexc |= ord(c) - ord('0')
                else:
                    hexc |= ord(c) - ord('A') + 0XA
                j+= 1
            rv+= chr(hexc)
            i+= j-1

        esc = False

    return rv



def p_stringValueList(p):
    """stringValueList : stringValue
                       | stringValueList stringValue
                       """
    if len(p) == 2:
        p[0] = _fixStringValue(p[1])
    else:
        p[0] = p[1] + _fixStringValue(p[2])


def p_constantValue(p):
    """constantValue : integerValue
                     | floatValue
                     | charValue
                     | stringValueList
                     | booleanValue
                     | nullValue
                     """
    p[0] = p[1]

def p_integerValue(p):
    """integerValue : binaryValue
                    | octalValue
                    | decimalValue
                    | hexValue
                    """
    p[0] = int(p[1])
    # TODO deal with non-decimal values. 

def p_referenceInitializer(p):
    """referenceInitializer : objectHandle
                            | aliasIdentifier
                            """
    if p[1][0] == '$':
        try:
            p[0] = p.parser.aliases[p[1]]
        except KeyError:
            ce = CIMError(CIM_ERR_FAILED, 
                    'Unknown alias: ' + p[0])
            ce.file_line = (p.parser.file, p.lexer.lineno)
            raise ce
    else:
        p[0] = p[1]

def p_objectHandle(p):
    """objectHandle : identifier"""
    p[0] = p[1]

def p_qualifierDeclaration(p):
    """qualifierDeclaration : QUALIFIER qualifierName qualifierType scope ';'
                            | QUALIFIER qualifierName qualifierType scope defaultFlavor ';'
                            """
    qualtype = p[3]
    dt, is_array, array_size, value = qualtype
    qualname = p[2]
    scopes = p[4]
    if len(p) == 5:
        flist = []
    else:
        flist = p[5]
    flavors = _build_flavors(flist)

    p[0] = CIMQualifierDeclaration(qualname, dt, value=value, 
                    is_array=is_array, array_size=array_size, 
                    scopes=scopes, **flavors)

def _build_flavors(flist, qualdecl=None):
    flavors = {}
    if qualdecl is not None:
        flavors = {'overridable':qualdecl.overridable,
                   'translatable':qualdecl.translatable,
                   'toinstance':qualdecl.toinstance,
                   'tosubclass':qualdecl.tosubclass}
    if 'disableoverride' in flist:
        flavors['overridable'] = False
    if 'enableoverride' in flist:
        flavors['overridable'] = True
    if 'translatable' in flist:
        flavors['translatable'] = True
    if 'restricted' in flist:
        flavors['tosubclass'] = False
    if 'tosubclass' in flist:
        flavors['tosubclass'] = True
    try:
        if flavors['tosubclass']:
            flavors['toinstance'] = True
    except KeyError:
        pass
    return flavors

def p_qualifierName(p):
    """qualifierName : identifier
                     | ASSOCIATION
                     | INDICATION
                     """
    p[0] = p[1]

def p_qualifierType(p):
    """qualifierType : qualifierType_1
                     | qualifierType_2
                     """
    p[0] = p[1]

def p_qualifierType_1(p):
    """qualifierType_1 : ':' dataType array
                       | ':' dataType array defaultValue
                       """
    dv = None
    if len(p) == 5:
        dv = p[4]
    p[0] = (p[2], True, p[3], dv)

def p_qualifierType_2(p):
    """qualifierType_2 : ':' dataType 
                       | ':' dataType defaultValue
                       """
    dv = None
    if len(p) == 4:
        dv = p[3]
    p[0] = (p[2], False, None, dv)

def p_scope(p):
    """scope : ',' SCOPE '(' metaElementList ')'"""
    slist = p[4]
    scopes = {}
    for i in ('SCHEMA',
              'CLASS',
              'ASSOCIATION',
              'INDICATION',
              'QUALIFIER',
              'PROPERTY',
              'REFERENCE',
              'METHOD',
              'PARAMETER',
              'ANY'):
        scopes[i] = i in slist
    p[0] = scopes

def p_metaElementList(p):
    """metaElementList : metaElement
                       | metaElementList ',' metaElement
                       """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[3]]

def p_metaElement(p):
    """metaElement : SCHEMA
                   | CLASS
                   | ASSOCIATION
                   | INDICATION
                   | QUALIFIER
                   | PROPERTY
                   | REFERENCE
                   | METHOD
                   | PARAMETER
                   | ANY
                   """
    p[0] = p[1].upper()

def p_defaultFlavor(p):
    """defaultFlavor : ',' FLAVOR '(' flavorListWithComma ')'"""
    flist = p[4]
    flavors = {'ENABLEOVERRIDE':True,
               'TOSUBCLASS':True,
               'DISABLEOVERRIDE':False,
               'RESTRICTED':False,
               'TRANSLATABLE':False}
    for i in flist:
        flavors[i] = True
    p[0] = flavors


def p_flavorListWithComma(p):
    """flavorListWithComma : flavor
                           | flavorListWithComma ',' flavor
                           """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[3]]

def p_instanceDeclaration(p):
    """instanceDeclaration : INSTANCE OF className '{' valueInitializerList '}' ';'
                           | INSTANCE OF className alias '{' valueInitializerList '}' ';'
                           | qualifierList INSTANCE OF className '{' valueInitializerList '}' ';'
                           | qualifierList INSTANCE OF className alias '{' valueInitializerList '}' ';'
                           """
    alias = None
    quals = {}
    ns = p.parser.handle.default_namespace
    if isinstance(p[1], basestring): # no qualifiers
        cname = p[3]
        if p[4] == '{':
            props = p[5]
        else:
            props = p[6]
            alias = p[4]
    else:
        cname = p[4]
        #quals = p[1] # qualifiers on instances are deprecated -- rightly so. 
        if p[5] == '{':
            props = p[6]
        else:
            props = p[7]
            alias = p[5]

    try:
        cc = p.parser.handle.GetClass(cname,
                LocalOnly=False, IncludeQualifiers=True)
        p.parser.classnames[ns].append(cc.classname.lower())
    except CIMError, ce:
        ce.file_line = (p.parser.file, p.lexer.lineno)
        if ce.args[0] == CIM_ERR_NOT_FOUND:
            file_ = p.parser.mofcomp.find_mof(cname)
            if p.parser.verbose:
                p.parser.log('Class %s does not exist' % cname)
            if file_:
                p.parser.mofcomp.compile_file(file_, ns)
                cc = p.parser.handle.GetClass(cname, LocalOnly=False, 
                        IncludeQualifiers=True)
            else:
                if p.parser.verbose:
                    p.parser.log("Can't find file to satisfy class")
                ce = CIMError(CIM_ERR_INVALID_CLASS, cname)
                ce.file_line = (p.parser.file, p.lexer.lineno)
                raise ce
        else:
            raise
    path = CIMInstanceName(cname, namespace=ns)
    inst = CIMInstance(cname, properties=cc.properties, 
            qualifiers=quals, path=path)
    for prop in props: 
        pname = prop[1]
        pval = prop[2]
        try:
            cprop = inst.properties[pname]
            cprop.value = tocimobj(cprop.type, pval)
        except KeyError:
            ce = CIMError(CIM_ERR_INVALID_PARAMETER, 
                    'Invalid property: %s' % pname)
            ce.file_line = (p.parser.file, p.lexer.lineno)
            raise ce
        except ValueError, ve:
            ce = CIMError(CIM_ERR_INVALID_PARAMETER, 
                    'Invalid value for property: %s: %s' % (pname,ve.message))
            ce.file_line = (p.parser.file, p.lexer.lineno)
            raise ce

    for prop in inst.properties.values():
        if 'key' not in prop.qualifiers or not prop.qualifiers['key']:
            continue
        if prop.value is None: 
            ce = CIMError(CIM_ERR_FAILED, 
                    'Key property %s.%s is not set' % (cname, prop.name))
            ce.file_line = (p.parser.file, p.lexer.lineno)
            raise ce
        inst.path.keybindings[prop.name] = prop.value

    if alias:
        p.parser.aliases[alias] = inst.path
    p[0] = inst 

def p_valueInitializerList(p):
    """valueInitializerList : valueInitializer
                            | valueInitializerList valueInitializer
                            """
    if len(p) == 2:
        p[0] = [p[1]]
    else:
        p[0] = p[1] + [p[2]]
    

def p_valueInitializer(p):
    """valueInitializer : identifier defaultValue ';'
                        | qualifierList identifier defaultValue ';'
                        """
    if len(p) == 4:
        id_ = p[1]
        val = p[2]
        quals = []
    else:
        quals = p[1]
        id_ = p[2]
        val = p[3]
    p[0] = (quals, id_, val)

def p_booleanValue(p):
    """booleanValue : FALSE
                    | TRUE
                    """
    p[0] = p[1].lower() == 'true'

def p_nullValue(p):
    """nullValue : NULL"""
    p[0] = None

def p_identifier(p):
    """identifier : IDENTIFIER
                  | ANY
                  | AS
                  | CLASS
                  | DISABLEOVERRIDE
                  | dataType 
                  | ENABLEOVERRIDE
                  | FLAVOR
                  | INSTANCE
                  | METHOD
                  | OF
                  | PARAMETER
                  | PRAGMA
                  | PROPERTY
                  | QUALIFIER
                  | REFERENCE
                  | RESTRICTED
                  | SCHEMA
                  | SCOPE
                  | TOSUBCLASS
                  | TRANSLATABLE
                  """
                  #| ASSOCIATION
                  #| INDICATION
    p[0] = p[1]

def p_empty(p):
    'empty :'
    pass

def _find_column(input, token):
    i = token.lexpos
    while i > 0:
        if input[i] == '\n':
            break
        i-= 1
    column = (token.lexpos - i)+1
    return column

def _get_error_context(input, token):
    try:
        line = input[token.lexpos : input.index('\n', token.lexpos)]
    except ValueError:
        line = input[token.lexpos:]
    i = input.rfind('\n', 0, token.lexpos)
    if i < 0: 
        i = 0
    line = input[i:token.lexpos] + line
    lines = [line.strip('\r\n')]
    col = token.lexpos - i
    while len(lines) < 5 and i > 0:
        end = i
        i = input.rfind('\n', 0, i)
        if i < 0:
            i = 0
        lines.insert(0, input[i:end].strip('\r\n'))
    pointer = ''
    for ch in token.value:
        pointer+= '^'
    pointline = ''
    i = 0
    while i < col -1:
        if lines[-1][i].isspace():
            pointline+= lines[-1][i]
            # otherwise, tabs complicate the alignment
        else:
            pointline+= ' '
        i+= 1
    lines.append(pointline + pointer)
    return lines

def _print_logger(str):
    print str


class MOFWBEMConnection(object):
    def __init__(self, conn=None):
        self.conn = conn
        self.class_names = {}
        self.qualifiers = {}
        self.instances = {}
        self.classes = {}
        if conn is None:
            self.__default_namespace = 'root/cimv2'

    def setns(self, value):
        if self.conn is not None:
            self.conn.default_namespace = value
        else:
            self.__default_namespace = value

    def getns(self):
        if self.conn is not None:
            return self.conn.default_namespace
        else:
            return self.__default_namespace

    default_namespace = property(getns, setns, None, 
            "default_namespace property")

    def GetClass(self, *args, **kwargs):
        cname = len(args) > 0 and args[0] or kwargs['ClassName']
        try:
            cc = self.classes[self.default_namespace][cname]
        except KeyError:
            if self.conn is None:
                ce = CIMError(CIM_ERR_NOT_FOUND, cname)
                raise ce
            cc = self.conn.GetClass(*args, **kwargs)
            try:
                self.classes[self.default_namespace][cc.classname] = cc
            except KeyError:
                self.classes[self.default_namespace] = \
                        NocaseDict({cc.classname:cc})
        if 'LocalOnly' in kwargs and not kwargs['LocalOnly']:
            if cc.superclass:
                try:
                    del kwargs['ClassName']
                except KeyError:
                    pass
                if len(args) > 0:
                    args = args[1:]
                super_ = self.GetClass(cc.superclass, *args, **kwargs)
                for prop in super_.properties.values():
                    if prop.name not in cc.properties:
                        cc.properties[prop.name] = prop
                for meth in super_.methods.values():
                    if meth.name not in cc.methods:
                        cc.methods[meth.name] = meth
        return cc

    def CreateClass(self, *args, **kwargs):
        cc = len(args) > 0 and args[0] or kwargs['NewClass']
        if cc.superclass:
            try:
                super_ = self.GetClass(cc.superclass, LocalOnly=True, 
                        IncludeQualifiers=False)
            except CIMError, ce:
                if ce.args[0] == CIM_ERR_NOT_FOUND:
                    ce.args = (CIM_ERR_INVALID_SUPERCLASS, cc.superclass)
                    raise
                else:
                    raise

        try:
            self.classes[self.default_namespace][cc.classname] = cc
        except KeyError:
            self.classes[self.default_namespace] = \
                        NocaseDict({cc.classname:cc})

        # TODO: should we see if it exists first with 
        # self.conn.GetClass()?  Do we want to create a class
        # that already existed? 
        try:
            self.class_names[self.default_namespace].append(cc.classname)
        except KeyError:
            self.class_names[self.default_namespace] = [cc.classname]

    def ModifyClass(self, *args, **kwargs):
        raise CIMError(CIM_ERR_FAILED, 
                'This should not happen!')

    def ModifyInstance(self, *args, **kwargs):
        raise CIMError(CIM_ERR_FAILED, 
                'This should not happen!')

    def GetQualifier(self, *args, **kwargs):
        qualname = len(args) > 0 and args[0] or kwargs['QualifierName']
        try:
            qual = self.qualifiers[self.default_namespace][qualname]
        except KeyError:
            if self.conn is None:
                raise CIMError(CIM_ERR_NOT_FOUND, qualname)
            qual = self.conn.GetQualifier(*args, **kwargs)
        return qual

    def SetQualifier(self, *args, **kwargs):
        qual = len(args) > 0 and args[0] or kwargs['QualifierDeclaration']
        try:
            self.qualifiers[self.default_namespace][qual.name] = qual
        except KeyError:
            self.qualifiers[self.default_namespace] = \
                    NocaseDict({qual.name:qual})

    def EnumerateQualifiers(self, *args, **kwargs):
        if self.conn is not None:
            rv = self.conn.EnumerateQualifiers(*args, 
                    **kwargs)
        else:
            rv = []
        try:
            rv+= self.qualifiers[self.default_namespace].values()
        except KeyError:
            pass
        return rv


    def CreateInstance(self, *args, **kwargs):
        inst = len(args) > 0 and args[0] or kwargs['NewInstance']
        try:
            self.instances[self.default_namespace].append(inst)
        except KeyError:
            self.instances[self.default_namespace] = [inst]
        return inst.path

    def rollback(self, verbose=False):
        for ns, insts in self.instances.items():
            insts.reverse()
            for inst in insts:
                try:
                    if verbose:
                        print 'Deleting instance', inst.path
                    self.conn.DeleteInstance(inst.path)
                except CIMError, ce:
                    print 'Error deleting instance', inst.path
                    print '    ', '%s %s' % (ce.args[0], ce.args[1])
        for ns, cnames in self.class_names.items():
            self.default_namespace = ns
            cnames.reverse()
            for cname in cnames:
                try:
                    if verbose:
                        print 'Deleting class %s:%s' % (ns, cname)
                    self.conn.DeleteClass(cname)
                except CIMError, ce:
                    print 'Error deleting class %s:%s' % (ns, cname)
                    print '    ', '%s %s' % (ce.args[0], ce.args[1])
        # TODO: do we want to do anything with qualifiers? 


def _errcode2string(code):
    d = {
        CIM_ERR_FAILED                 : 'A general error occurred',
        CIM_ERR_ACCESS_DENIED          : 'Resource not available',
        CIM_ERR_INVALID_NAMESPACE      : 'The target namespace does not exist',
        CIM_ERR_INVALID_PARAMETER      : 'Parameter value(s) invalid',
        CIM_ERR_INVALID_CLASS          : 'The specified Class does not exist',
        CIM_ERR_NOT_FOUND              : 'Requested object could not be found',
        CIM_ERR_NOT_SUPPORTED          : 'Operation not supported',
        CIM_ERR_CLASS_HAS_CHILDREN     : 'Class has subclasses',
        CIM_ERR_CLASS_HAS_INSTANCES    : 'Class has instances',
        CIM_ERR_INVALID_SUPERCLASS     : 'Superclass does not exist',
        CIM_ERR_ALREADY_EXISTS         : 'Object already exists',
        CIM_ERR_NO_SUCH_PROPERTY       : 'Property does not exist',
        CIM_ERR_TYPE_MISMATCH          : 'Value incompatible with type',
        CIM_ERR_QUERY_LANGUAGE_NOT_SUPPORTED   : 'Query language not supported',
        CIM_ERR_INVALID_QUERY          : 'Query not valid',
        CIM_ERR_METHOD_NOT_AVAILABLE   : 'Extrinsic method not executed',
        CIM_ERR_METHOD_NOT_FOUND       : 'Extrinsic method does not exist',
        }
    try:
        s = d[code]
    except KeyError:
        s = 'Unknown Error'
    return s

class MOFCompiler(object):
    def __init__(self, handle, search_paths=[], verbose=False,
                 log_func=_print_logger):
        """Initialize the compiler.

        Keyword arguments:
        handle -- A WBEMConnection or similar object.  The following 
            attributes and methods need to be present, corresponding to the 
            the attributes and methods on pywbem.WBEMConnection having 
            the same names:
            - default_namespace
            - EnumerateInstanceNames()
            - CreateClass()
            - GetClass()
            - ModifyClass()
            - DeleteInstance()
            - CreateInstance()
            - ModifyInstance()
            - DeleteQualifier()
            - EnumerateQualifiers()
            - SetQualifier()
        search_paths -- A list of file system paths specifying where 
            missing schema elements should be looked for. 
        verbose -- True if extra messages should be printed. 
        log_func -- A callable that takes a single string argument.  
            The default logger prints to stdout. 
        """

        self.parser = yacc.yacc(tabmodule=_tabmodule, optimize=_optimize)
        self.parser.search_paths = search_paths
        self.handle = handle
        self.parser.handle = handle
        self.lexer = lex.lex(lextab=_lextab, optimize=_optimize)
        self.lexer.parser = self.parser
        self.parser.qualcache = {handle.default_namespace:NocaseDict()}
        self.parser.classnames = {handle.default_namespace:[]}
        self.parser.mofcomp = self
        self.parser.verbose = verbose
        self.parser.log = log_func
        self.parser.aliases = {}

    def compile_string(self, mof, ns, filename=None):
        """Compile a string of MOF.

        Arguments:
        mof -- The string of MOF
        ns -- The CIM namespace

        Keyword arguments:
        filename -- The name of the file that the MOF was read from.  This 
            is used in status and error messages.
        """

        lexer = self.lexer.clone()
        lexer.parser = self.parser
        try:
            oldfile = self.parser.file
        except AttributeError:
            oldfile = None
        self.parser.file = filename
        try:
            oldmof = self.parser.mof
        except AttributeError:
            oldmof = None
        self.parser.mof = mof
        self.parser.handle.default_namespace = ns
        if ns not in self.parser.qualcache:
            self.parser.qualcache[ns] = NocaseDict()
        if ns not in self.parser.classnames:
            self.parser.classnames[ns] = []
        try:
            rv = self.parser.parse(mof, lexer=lexer)
            self.parser.file = oldfile
            self.parser.mof = oldmof
            return rv
        except MOFParseError, pe:
            self.parser.log('Syntax error:')
            if hasattr(pe, 'file') and hasattr(pe, 'lineno'):
                self.parser.log('%s:%s:' % (pe.file, pe.lineno))
            if hasattr(pe, 'context'):
                self.parser.log('\n'.join(pe.context))
            if str(pe):
                self.parser.log(str(pe))
            raise
        except CIMError, ce:
            if hasattr(ce, 'file_line'):
                self.parser.log('Fatal Error: %s:%s' % (ce.file_line[0], 
                                                        ce.file_line[1]))
            else:
                self.parser.log('Fatal Error:')
            self.parser.log('%s%s' % (_errcode2string(ce.args[0]), 
                    ce.args[1] and ': '+ce.args[1] or ''))
            raise

    def compile_file(self, filename, ns):
        """Compile MOF from a file.

        Arguments:
        filename -- The file to read MOF from
        ns -- The CIM namespace
        """

        if self.parser.verbose:
            self.parser.log('Compiling file ' + filename)
        f = open(filename, 'r')
        mof = f.read()
        f.close()

        return self.compile_string(mof, ns, filename=filename)

    def find_mof(self, classname):
        """Find a MOF file corresponding to a CIM class name.  The search_paths
        provided to __init__() are searched recursively.

        Arguments:
        classname -- The name of the class to look for
        """

        classname = classname.lower()
        for search in self.parser.search_paths:
            for root, dirs, files in os.walk(search):
                for file_ in files:
                    if file_.endswith('.mof') and \
                            file_[:-4].lower() == classname:
                        return root + '/' + file_
        return None

    def rollback(self, verbose=False):
        self.handle.rollback(verbose=verbose)

def _build():
    yacc.yacc(optimize=_optimize, tabmodule=_tabmodule)
    lex.lex(optimize=_optimize, lextab=_lextab)


if __name__ == '__main__':
    from optparse import OptionParser
    usage = 'usage: %prog -n <namespace> [options] <MOF file> ...'
    oparser = OptionParser(usage=usage)
    oparser.add_option('-s', '--search', dest='search', 
            help='Search path to find missing schema elements.  This option can be present multiple times.', 
            metavar='Path', action='append')
    oparser.add_option('-n', '--namespace', dest='ns', 
            help='Specify the namespace', metavar='Namespace')
    oparser.add_option('-u', '--url', dest='url', 
            help='URL to the CIM Server', metavar='URL', 
            default='/var/run/tog-pegasus/cimxml.socket')
    oparser.add_option('-v', '--verbose',
            action='store_true', dest='verbose', default=False,
            help='Print more messages to stdout')
    oparser.add_option('-r', '--remove',
            action='store_true', dest='remove', default=False,
            help='Remove elements found in MOF, instead of create them')
    oparser.add_option('-l', '--username',
            dest='username', metavar='Username',
            help='Specify the username')
    oparser.add_option('-p', '--password', 
            dest='password', metavar='Password',
            help='Specify the password')

    (options, args) = oparser.parse_args()
    search = options.search
    if not args:
        oparser.error('No input files given for parsing')
    if options.ns is None: 
        oparser.error('No namespace given')

    passwd = options.password
    if options.username and not passwd:
        passwd = getpass('Enter password for %s: ' % options.username)
    if options.username:
        conn = WBEMConnection(options.url, (options.username, passwd))
    else:
        conn = WBEMConnection(options.url)
    if options.remove:
        conn = MOFWBEMConnection(conn=conn)
    #conn.debug = True
    conn.default_namespace = options.ns
    if search is None:
        search = []

    search = [os.path.abspath(x) for x in search]
    for fname in args:
        path = os.path.abspath(os.path.dirname(fname))
        for spath in search:
            if path.startswith(spath):
                break
        else:
            search.append(path)

    # if removing, we'll be verbose later when we actually remove stuff.
    # We don't want MOFCompiler to be verbose, as that would be confusing. 
    verbose = options.verbose and not options.remove

    mofcomp = MOFCompiler(handle=conn, search_paths=search, 
            verbose=verbose)

    try:
        for fname in args:
            if fname[0] != '/':
                fname = os.path.curdir + '/' + fname
            mofcomp.compile_file(fname, options.ns)
    except MOFParseError, pe:
        sys.exit(1)
    except CIMError, ce:
        sys.exit(1)

    if options.remove:
        conn.rollback(verbose=options.verbose)





#
# mkbackup/install.py
#
# Windows, Linux and ghettoVCB.sh installation 


import os, sys, time, re
import platform, subprocess
import urllib2, tarfile, tempfile, StringIO, shutil

sample_ini_url='http://www.magikmon.com/download/mksbackup/sample.ini'
ghettovcb_url='http://www.magikmon.com/download/mksbackup/ghettoVCB-20110628.tar.gz'
readme_txt_url='http://www.magikmon.com/download/mksbackup/README.txt'
#ftpput_url='http://www.magikmon.com/download/mksbackup/ftpput'


# =====================================================================
def raw_input_with_default(msg, prompt, default):

#    try:
#        import pyreadline as readline
#            
#        def pre_input_hook():
#            readline.insert_text(default)
#            readline.redisplay()
#        
#        readline.set_pre_input_hook(pre_input_hook)
#        try:
#            return raw_input(msg+prompt)
#        finally:
#            readline.set_pre_input_hook(None)
#    except ImportError, e:
#        print 'ERROR', e
        if default:
            line="%s [%s]%s" % (msg, default, prompt)
        else:
            line="%s%s" % (msg, prompt)
        value=raw_input(line)
        if value=='' and default:
            return default
        return value

# =====================================================================
def Install():
    """Install"""
    if sys.platform in ('win32', ):
        Windows_Install()
    else:
        Linux_Install()
        
# =====================================================================
def Install_GhettoVCB(mksbackup_target, sources=None):
        print 'Download ghettoVCB.sh (last at http://communities.vmware.com/docs/DOC-8760) '
        vmware_dir=os.path.join(mksbackup_target, 'vmware')
        try:
            os.mkdir(vmware_dir)
        except Exception:
            pass
        try:
            data=urllib2.urlopen(ghettovcb_url, None).read()
        except (urllib2.URLError, urllib2.HTTPError, Exception), e:
            print 'Download failed'
        else:
            print 'Install ghettoVCB.sh and ghettoVCB.conf in sub directory vmware'
            cwd=os.getcwd()
            tmp_dir=tempfile.mkdtemp()
            os.chdir(tmp_dir)
            try:
                tarfile.open(fileobj=StringIO.StringIO(data), mode='r:gz').extractall()
            except Exception, e:
                print 'Unexpected error extraction ghettoVCB.tar.gz (%s)' % (e,)
                os.chdir(cwd)
            else:
                os.chdir(cwd)
                for filename, rename in [ ('ghettoVCB.conf', 'ghettoVCB.conf.new'), ('ghettoVCB.sh', None), ('ghettoVCB-vm_backup_configuration_template', None), ]:
                    dst_filename=os.path.join(vmware_dir, filename)
                    src_filename=os.path.join(tmp_dir, 'ghettoVCB', filename)
                    if rename and os.path.isfile(dst_filename):
                        dst_filename=os.path.join(vmware_dir, rename)
                    
                    if os.path.isfile(src_filename):
                        data=open(src_filename).read() 
                        if sys.platform in ('win32', ) and filename in ('ghettoVCB.conf','ghettoVCB-vm_backup_configuration_template', ):
                            # convert text file in windows format
                            data=re.sub('(?<!\r)\n','\r\n', data)
                        f=open(dst_filename, 'wb')
                        f.write(data)
                        f.close()
                    else:
                        print 'file already exists: ' + src_filename

#        print 'Download ftpput a ftp uploader for VMware'
#        try:
#            data=urllib2.urlopen(ftpput_url, None).read()
#        except (urllib2.URLError, urllib2.HTTPError, Exception), e:
#            print 'Download failed (%s)' % (e, )
#        else:
#            dst_filename=os.path.join(vmware_dir, 'ftpput')
#            open(dst_filename, 'w').write(data)
    

# =====================================================================
def Install_ini(source_dir, target_dir):
    if source_dir:
        sample_ini=os.path.join(source_dir, 'sample.ini')
        
    if source_dir and os.path.isfile(sample_ini):
        sample_ini_data=open(sample_ini).read()
    else:
        print 'Retrieve sample.ini file from %s ...' % (sample_ini_url, )
        try:
            sample_ini_data=urllib2.urlopen(sample_ini_url, None).read()
        except (urllib2.URLError, urllib2.HTTPError, Exception), e:
            sample_ini_data='# mksbackup.ini more on http://www.magikmon.com/mksbackup\n# sample .ini file at %s\n' % (sample_ini_url, )
            pass

    if sys.platform in ('win32', ):
        # convert text file in windows format 
        sample_ini_data=re.sub('(?<!\r)\n','\r\n', sample_ini_data)
    
    mksbackup_ini=os.path.join(target_dir, 'mksbackup.ini')
    if os.path.isfile(mksbackup_ini):
        print 'Config file already exists: %s' % (mksbackup_ini, )
    else:
        print 'Create sample config file: %s' % (mksbackup_ini, )
        open(mksbackup_ini, 'wb').write(sample_ini_data)
        
    return mksbackup_ini

# =====================================================================
def Linux_Install():
    """create directories in /etc and download ghettoVCB"""

    mksbackup_target='/etc/mksbackup'
    choice=raw_input_with_default('Do you want to setup config files in %s ? (y/n)' % (mksbackup_target, ), '>', 'y')
    if choice!='y':
        print 'bye'
        return 0
        
    if not os.path.isdir(mksbackup_target):
        os.makedirs(mksbackup_target)
    
    Install_ini(None, mksbackup_target)
       
    if os.path.isdir('/etc/cron.d'):
        if not os.path.isfile('/etc/cron.d/mksbackup'):
            mksbackup_starter='"%s" "%s"' % (sys.executable, os.path.abspath(sys.argv[0]))
            cron_file=open('/etc/cron.d/mksbackup', 'w')
            print >>cron_file, '# MKSBACKUP cron file http://www.magikmon.com/mksbackup'
            print >>cron_file, '# un-comment and edit line bellow'
            print >>cron_file, '# 45 22 * * * root %s -c /etc/mksbackup/mksbackup.ini backup BACKUP_JOB' % mksbackup_starter
            cron_file.close()
        else:
            print 'file already exists: /etc/cron.d/mksbackup'

    choice=raw_input_with_default('Do you want to download and install ghettoVCB ?', '>', 'y')
    if choice=='y':
        Install_GhettoVCB(mksbackup_target)

    print 'More at http://www.magikmon.com/mksbackup\n' 
    print 'Edit and modify files: /etc/mksbackup/mksbackup.ini and /etc/cron.d/mksbackup'


# =====================================================================
def Windows_Install():
    """Install the exe, setup a default .ini and create task in scheduler"""

    choice=raw_input_with_default('Do you want to install MKSBackup ? (y/n)', '>', 'y')
    if choice!='y':
        print 'bye'
        return 0
        
    target=raw_input_with_default('Installation directory ', '>', 'C:\Magik')
    mksbackup_target=os.path.join(target, 'MKSBackup')
    if os.path.isdir(mksbackup_target):
        print 'Directory %s already exists, continue.' % (mksbackup_target,)
    else:
        try:
            os.makedirs(mksbackup_target)
        except Exception, e:
            print 'Error creating directory %s (%s)', (mksbackup_target, e)
        else:
            print 'Directory created'

    import shutil
    
    source=os.path.dirname(sys.argv[0])
    # copy .exe
    if sys.argv[0][-4:].lower()=='.exe':
        if os.path.realpath(sys.argv[0])==os.path.realpath(mksbackup_target):
            """.exe already in place"""
        else:
            print 'Copy %s' % (os.path.basename(sys.argv[0]))
            shutil.copy(sys.argv[0], mksbackup_target)

    # copy and rename sample.ini
    doc_src=os.path.join(source, 'doc')
    mksbackup_ini=Install_ini(doc_src, target)
        
    # copy doc
    if os.path.isdir(doc_src):
        doc_dst=os.path.join(mksbackup_target, 'doc')
        try:
            os.mkdir(doc_dst)
        except Exception:
            pass
        try:
            for filename in os.listdir(doc_src):
                shutil.copy(os.path.join(doc_src, filename), doc_dst)
        except Exception, e:
            print 'Error copying documentation, continue'
        else:
            print 'Copy documentation'
    else:
        # try to get Readme
        print 'Documentation not found, try to retrieve Readme.txt'
        print 'from %s ...' % (readme_txt_url, )
        try:
            readme_txt_data=urllib2.urlopen(readme_txt_url, None).read()
        except (urllib2.URLError, urllib2.HTTPError, Exception), e:
            readme_txt_data='more on http://www.magikmon.com/mksbackup\n'
        open(os.path.join(mksbackup_target, 'Readme.txt'), 'w').write(readme_txt_data)

    # install ghettoVCB and other VMware tools
    ghettovcb_choice=raw_input_with_default('Do you want to download and install ghettoVCB ?', '>', 'y')
    if ghettovcb_choice=='y':
        Install_GhettoVCB(target)
        
    # create task in scheduler if not already exists
    stdout, stderr=subprocess.Popen('SCHTASKS /Query', stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
    if stdout.find('MKSBackup')==-1:
        print 'create task MKSBackup in scheduler'
        jobs=raw_input_with_default('Enter your job names separated by spaces', '>', 'BACKUP_JOB')
        cmd='"%s" -q -l "%s" -c "%s" backup %s' % (os.path.join(mksbackup_target, os.path.basename(sys.argv[0])), os.path.join(target, 'mksbackup.log'), os.path.join(target, 'mksbackup.ini'), jobs) 
        schtasks_cmd=[ 'SCHTASKS', '/Create', '/SC', 'DAILY', '/TN', 'MKSBackup', '/ST', '22:45:00', '/RU', os.getenv('USERNAME'),  '/TR', cmd ]
        if sys.getwindowsversion()[0]>5:
            # under 2008 backup require HIGHEST privilege
            schtasks_cmd.insert(2, 'HIGHEST')
            schtasks_cmd.insert(2, '/RL')
            # under 2008, to force the system to ask for the password, set empty password 
            i=schtasks_cmd.index('/RU')+2
            schtasks_cmd.insert(i, '')
            schtasks_cmd.insert(i, '/RP')
        else:
            pass
            
        # print ' '.join(schtasks_cmd)
        process=subprocess.Popen(schtasks_cmd)
        _code=process.wait()
    else:
        print 'task MKSBackup already exists, skip'

    print 'Open task scheduler'
    subprocess.Popen([ 'control.exe',  'schedtasks', ])  # taskschd.msc works under 2008 not on XP
    print 'Open config file in notepad'    
    subprocess.Popen([ 'notepad', mksbackup_ini] )
    if ghettovcb_choice=='yes':
        print 'Open ghettoVCB.conf file in notepad', os.path.join(target, 'vmware\ghettoVCB.conf')    
        subprocess.Popen([ 'notepad', os.path.join(target, 'vmware\ghettoVCB.conf') ] )
        
    print 'Update or review your configuration and task'
    print 'More at http://www.magikmon.com/mksbackup' 
    raw_input('Install completed, press ENTER to quit')

    return 0


#
# archiver.py
#

import os, sys, socket, re, time, smtplib, calendar, StringIO, logging, logging.handlers, locale, gzip, posixpath
import urllib, urllib2, urlparse, ftplib
from datetime import datetime, timedelta
import email
from email.Utils import parseaddr, formataddr, formatdate
from email.MIMEMultipart import MIMEMultipart
from email.MIMEText import MIMEText
from email.MIMEBase import MIMEBase

import cron

atom=r"[a-zA-Z0-9_#\$&'*+/=?\^`{}~|\-]+"
dot_atom=atom  +  r"(?:\."  +  atom  +  ")*"
quoted=r'"(?:\\[^\r\n]|[^\\"])*"'
local="(?:"  +  dot_atom  +  "|"  +  quoted  +  ")"
domain_lit=r"\[(?:\\\S|[\x21-\x5a\x5e-\x7e])*\]"
domain="(?:"  +  dot_atom  +  "|"  +  domain_lit  +  ")"
addr_spec=local  +  "\@"  +  domain
postfix_restricted_rfc2822_address_name=local
postfix_restricted_rfc2822_email_address=addr_spec
cyrus_mailbox_name=r"[a-zA-Z0-9_#$'=`{}~|-]+(?:\.[a-zA-Z0-9_#$'=`{}~|-]+)*"

domain_nameRE=re.compile('^'+dot_atom+'$')
email_addressRE=re.compile('^'+postfix_restricted_rfc2822_email_address+'$')
valid_ipRE=re.compile('^([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\\.([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$')
valid_hostnameRE=re.compile('^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\\-]*[a-zA-Z0-9])\\.)*([A-Za-z]|[A-Za-z][A-Za-z0-9\\-]*[A-Za-z0-9])$')
ftp_url_re=re.compile(r'ftp://(?:(?P<username>[\w\.\-\+%!$&\'\(\)*,;=#]+)(?::(?P<password>[\w\.\-\+%!$&\'\(\)*,;=#]+))?@)?(?P<host>[a-zA-Z0-9\.\-]+)(?::(?P<port>\d+))?(?:(?P<directory>/.*))?')

# from FormEncode
url_re = re.compile(r'''
    ^(http|https)://
    (?:[%:\w]*@)?                           # authenticator
    (?P<domain>([a-z0-9][a-z0-9\-]{1,62}\.)*) # (sub)domain - alpha followed by 62max chars (63 total)
    (?P<tld>[a-z]{2,})                      # TLD
    (?::[0-9]+)?                            # port
    # files/delims/etc
    (?P<path>/[a-z0-9\-\._~:/\?#\[\]@!%\$&\'\(\)\*\+,;=]*)?
    $
''', re.I | re.VERBOSE)

boolean={ 'on':True, 'yes':True, 'true':True, '1':True, 1:True, 'off':False, 'no':False, 'false':False, '0':False, 0:False}

# ---------------------------------------------------------------------
def check_boolean(job, errors, options):
    for option, default in options:
        value=boolean.get(job.get(option, default).lower(), None)
        if value==None:
            errors[option]='boolean must have value in (on, yes, true, 1, off, no, false and 0)'
            value=default
        job[option]=value


# ---------------------------------------------------------------------------
def check_mail_config(job):
    
    errors={}

    defaul_port=dict(normal=25, ssl=465, tls=587)
    
    mail=boolean.get(job.get('mail', 'yes'), None)
    if mail==None:
        errors['mail']='boolean must have value in (on, yes, true, 1, off, no, false and 0)'
    else:
        job['mail']=mail

    magikmon=job.get('magikmon', '')
    if magikmon:
        if not url_re.match(magikmon):
            errors['magikmon']='not a valid URL'
        else:
            if job['program']=='esxmon':
                up=urlparse.urlparse(magikmon)
                if up[1].lower().startswith('monitoring.magikmon.com') and up[2].lower().startswith('/sendbackup'):
                    magikmon=''
                    # don't send esxmon status to MagiKmon backup service 
                    # errors['magikmon']='Cannot report status to MagikMon backup service'
                    
    job['magikmon']=magikmon
    
    if mail:
        try:
            job['smtp_host']=smtp_host=job['smtp_host'].encode('ascii')
        except UnicodeEncodeError:
            errors['smtp_host']='invalid hostname or ip address'
        else:
            if not valid_ipRE.match(smtp_host):
                if not valid_hostnameRE.match(smtp_host):
                    errors['smtp_host']='invalid hostname or ip address'
                else:
                    try:
                        ip=socket.gethostbyname(smtp_host)
                    except socket.gaierror:
                        errors['smtp_host']='cannot resolve address'
    
        smtp_mode=job['smtp_mode'].lower()
        if not smtp_mode in ('normal', 'ssl', 'tls'):
            errors['smtp_mode']='mode must be in "normal", "ssl", or "tls"'
        else:
            job['smtp_mode']=smtp_mode
            if smtp_mode in ('ssl', 'tls') and not hasattr(socket, 'ssl'):
                errors['smtp_mode']='this version of python has not support for SSL'
            if smtp_mode=='ssl' and not hasattr(smtplib, 'SMTP_SSL'):
                errors['smtp_mode']='this version of python has not SSL support for SMTP'
                
        job['smtp_login']=job.get('smtp_login')
        job['smtp_password']=job.get('smtp_password')

        smtp_port=job.get('smtp_port', defaul_port.get(smtp_mode, 25))
        try:
            port=int(smtp_port)
        except Exception:
            errors['smtp_port']='must be an integer'
        else:
            if not (0<port and port<65535):
                errors['smtp_port']='must be an integer between 1 and 65535'
            else:
                job['smtp_port']=port
            
    if mail or magikmon:

        sender=job.get('sender', None)
        if not sender:
            errors['sender']='option mandatory'
        elif not email_addressRE.match(sender):
            errors['sender']='invalid email address'
    
        recipients=job.get('recipients')
        if not recipients:
            errors['recipients']='option mandatory'
        else:
            recipients=recipients.split()
            bad=[]
            for recipient in recipients[:]:
                if not email_addressRE.match(recipient):
                    bad.append(recipient)
                domain=recipient.split('@')[-1].lower()
                if job['program']=='esxmon' and  domain=='monitoring.magikmon.com':
                    # don't send esxmon status to MagiKmon backup service 
                    # errors['recipients']='Cannot report status to MagikMon backup service'
                    recipients.remove(recipient)
                    
            if bad:
                errors['recipients']='invalid email addresses: %s' % ' '.join(bad)
            else:
                job['recipients']=recipients

        attachment_gzip=job['attachment_gzip']
        try:
            size=int(attachment_gzip)
        except Exception:
            errors['attachment_gzip']='must be an integer'
        else:
            if not (0<size and size<=10240):
                errors['attachment_gzip']='must be an integer between 1 and 10240'
            else:
                job['attachment_gzip']=size*1024 # in K

        attachment_size=job['attachment_size']
        try:
            size=int(attachment_size)
        except Exception:
            errors['attachment_size']='must be an integer'
        else:
            if not (0<size and size<=10240):
                errors['attachment_size']='must be an integer between 1 and 10240'
            else:
                job['attachment_size']=size*1024 # in K


    return errors

# ---------------------------------------------------------------------------
def socket_readlines(chan):
    """iterator reading lines from a socket"""
    line_buf=''
    while True:
        x=chan.recv(1024)
        if len(x)==0:
            break
        line_buf=line_buf+x
        lines=line_buf.split('\n')
        line_buf=lines[-1]
        for line in lines[:-1]:
            yield line

    if line_buf:
        yield line_buf

# ---------------------------------------------------------------------------
def ftp_pre_create_dir(host, port, username, password, directory):
    ftpcli=ftplib.FTP()
    try:
        ftpcli.connect(host, port, timeout=10)
    except TypeError:
        # python 2.5 and below have no timeout
        ftpcli=ftplib.FTP()
        ftpcli.connect(host, port)
           
    ftpcli.login(username, password)
    dir=directory
    dirs=[]
    while dir and dir!='/':
        dirs.insert(0, dir)
        dir=posixpath.dirname(dir)

    for dir in dirs:
        try:
            ftpcli.mkd(dir)
        except:
            pass
    try:
        ftpcli.quit()
    except:
        pass
    ftpcli.close()

# ---------------------------------------------------------------------------
def _list_dir(target_dir, log):
    total_size=0
    dir_out=u'== directory %s\r\n  epoch   |          time          |      size     |    filename\r\n' % target_dir
    for filename in os.listdir(target_dir): # if target_dir is unicode, return os.listdir return unicode
        full_filename=os.path.join(target_dir, filename)
        stat=os.stat(full_filename)
        if os.path.isdir(full_filename):
            size='<DIR>'
        else:
            size='%15d' % stat.st_size
            total_size+=stat.st_size
        dir_out+='%d %s %15s %s\r\n' % (stat.st_mtime, time.ctime(stat.st_mtime), size, filename)
        
    return dir_out, total_size

# ---------------------------------------------------------------------------
def list_dir(target_dir, log):
    try:
        dir_out, total_size=_list_dir(target_dir, log) 
    except Exception, e:
        log.error('listing target directory: %s', e)
        dir_out='error listing directory "%s": %s\r\n' % (target_dir, e)
        total_size=0
        
    return dir_out, total_size

# ---------------------------------------------------------------------------
def free_space(target_dir, log):
    
    try:
        if sys.platform=='win32':
            import win32file
            free_bytes, total_bytes, total_free_bytes=win32file.GetDiskFreeSpaceEx(target_dir)
        else:
            import os, statvfs
            s=os.statvfs(target_dir)
            total_free_bytes=s[statvfs.F_BSIZE]*s[statvfs.F_BAVAIL] 

    except Exception, e:
        if log:
            log.error('checking directory "%s": %s', target_dir, e)
        raise e

    return total_free_bytes

# ---------------------------------------------------------------------------
def quoted_string_list(st):
    """
    st  is a list of string separated by spaces like in shell script
    String containing space can be quoted
    Quote can be escaped with \
    r'1 two "t h r e e" "with a \" inside"' -->
                        ['1', 'two', 't h r e e', 'with a \\" inside']
    '" dont remove this line my editor string highlighting need it
    """                        
    lst=[]
    s=st.strip()
    while s:
        match=re.match(r'("(?:\\"|[^"])*"|[^ "]*)(.*)',  s)
        if match:
            item=match.groups()[0]
            if not item:
                raise ValueError, ("not a quoted string list '%s'" % (st,))
            elif item[0]=='"':
                lst.append(item[1:-1])
            else:
                lst.append(item)
            s=match.groups()[1].lstrip()
        else:
            raise ValueError, ("not a quoted string list '%s'" % (st,))
    return lst

# ---------------------------------------------------------------------------
def sendmail(job, subject, text, attachments, log):

    if not job['mail'] and not job['magikmon']:
        return None

    sender=job['sender']
    recipients=job['recipients']
    attachment_size=job['attachment_size']
    attachment_gzip=job['attachment_gzip']

    msg=MIMEMultipart()
    msg.preamble='' # This line is not visible on mime enable MUA
    msg.epilogue=''
    
    msg['From'] = formataddr((sender, sender)) # Display name, email address
    msg['To'] =  ', '.join([ formataddr((recipient, recipient)) for recipient in recipients ])
    msg['Date'] = formatdate(localtime=True)
    msg['Subject'] = subject

    core=MIMEText(text.encode('utf-8'), 'plain', 'utf-8')
    msg.attach(core)
    
    for filename, target, data, maintype, subtype, charset in attachments:
        
        if not data and target and os.path.exists(target):
            data=open(target, 'rb').read(attachment_size)

        if data==None:
            data=''

        if len(data)>attachment_gzip:
            log.info('gzip attachment %s', filename)
            file_out=StringIO.StringIO()
            zfile=gzip.GzipFile(filename, 'wb', 9, file_out)
            if maintype=='text':
                # all txt gzip file are encode in UTF8 to avoid the loose of the charset 
                data=data.decode(charset).encode('utf-8')
            zfile.write(data)
            zfile.close()
            data=file_out.getvalue()
            file_out.close()
            filename+='.gz'
            maintype, subtype='application', 'x-gzip'

        if len(data)>attachment_size:
            log.warning('attachment is too big, skipped: %s', filename)
            # replace attachment by an "error"
            data='attachment is too big: %s\n' % (filename, )
            filename=os.path.splitext(filename)[0]+'-error.txt'
            maintype, subtype, charset='text', 'plain', 'ascii'
            
        #maintype, subtype = 'application', 'octet-stream'
        if maintype=='text':
            attachment=MIMEText(data, subtype, charset)
        else:
            attachment=MIMEBase(maintype, subtype)
            attachment.set_payload(data) #, charset) #dont use the charset, here !
            email.Encoders.encode_base64(attachment)

        attachment.add_header('Content-Disposition', 'attachment', filename=filename)
        msg.attach(attachment)
    
    body=msg.as_string()

    # first try send by http             
    if job['magikmon'] and job['program']!='esxmon':
        try:
            data=urllib.urlencode(dict(body=body))
            try:
                urllib2.urlopen(job['magikmon'], data, timeout=30)
            except TypeError:
                # python 2.4
                urllib2.urlopen(job['magikmon'], data)
        except Exception, e:
            log.exception('error sending mail to MagiKmon url: %s (%s)', job['magikmon'], e)
            if not errmsg:
                errmsg='error sending mail to MagiKmon url'
        else:
            log.debug('mail sent to MagiKmon url: %s', job['magikmon'])
            # Don't send the mail twice via HTTP and SMTP to the same MagiKmon monitoring service
            # remove the redundant recipient if any
            # http://monitoring.magikmon.com/sendbackup/XX/secret
            # b.XX@monitoring.magikmon.com
            #    
            domain, tld, path=url_re.match(job['magikmon']).group('domain', 'tld', 'path')
            match=re.match('/sendbackup/(?P<service_id>[0-9]+)(/.*)?', path)
            if match:
                service_id=int(match.group('service_id'))
                mkm_address='b.%d@%s%s' % (service_id, domain, tld)
                for recipient in recipients[:]:
                    if recipient.lower()==mkm_address.lower():
                        log.info('"%s" removed from the list of recipients', recipient)
                        recipients.remove(recipient)
                
    errmsg=''     
    if job['mail'] and recipients:
        smtp_host=job['smtp_host']
        smtp_port=job['smtp_port']
        smtp_mode=job['smtp_mode']
        smtp_login=job['smtp_login']
        smtp_password=job['smtp_password']
    
        rcpt=[]
        try:
            if smtp_mode=='ssl':
                smtp=smtplib.SMTP_SSL(smtp_host, smtp_port)
            else:
                smtp=smtplib.SMTP(smtp_host, smtp_port)
                if smtp_mode=='tls':
                    smtp.starttls()
                    
            if smtp_login and smtp_password:
                # login and password must be encoded because HMAC used in CRAM_MD5 require non unicode string
                smtp.login(smtp_login.encode('utf-8'), smtp_password.encode('utf-8'))
    
            ret=smtp.sendmail(sender, recipients, body)
            smtp.quit()
        except (socket.error, ), e:
            errmsg='server %s:%s not responding: %s' % (smtp_host, smtp_port, e)
        except smtplib.SMTPAuthenticationError, e:
            errmsg='authentication error: %s' % (e, )
        except smtplib.SMTPRecipientsRefused, e:
            # code, errmsg=e.recipients[recipient_addr]
            errmsg='recipients refused: '+', '.join(e.recipients.keys())
        except smtplib.SMTPSenderRefused, e:
            # e.sender, e.smtp_code, e.smtp_error
            errmsg='sender refused: %s' % (e.sender, )
        except smtplib.SMTPDataError, e:
            errmsg='SMTP protocol mismatch: %s' % (e, )
        except smtplib.SMTPHeloError, e:
            errmsg="server didn't reply properly to the HELO greeting: %s" % (e, )
        except smtplib.SMTPException, e:
            errmsg='SMTP error: %s' % (e, )
        except Exception, e:
            errmsg=str(e)
            log.exception('unexpected error sending mail: %s', e)
        else:
            rcpt=recipients[:]
            if ret:
                errmsg='recipients refused: '+', '.join(ret.keys())        
                for r in ret.keys():
                    rcpt.remove(r)
    
            if log:
                log.info('mail sent to: %s', ', '.join(rcpt))
    
        if log:
            if errmsg:
                log.error('sending mail: %s', errmsg)

    return errmsg


# ---------------------------------------------------------------------------
def send_mail_report(job, result, manager):
    if not job['mail'] and not job['magikmon']:
        return None
    
    backup_status, status, attachments=result

    msg_body=job['msg_header']+'\n'+status
    subject='MKSBACKUP BACKUP %s %s' % (backup_status, job['name'])
    
    attachments.extend(manager.attachments)
    attachments.append(('status.txt', None, status.encode('utf-8'), 'text', 'plain', 'utf-8'))
    log_output=manager.stop_logging()
    attachments.append(('logging.txt', None, log_output.encode('utf-8'), 'text', 'plain', 'utf-8'))                       
    
    for line in status.split('\r\n'):
        if line:
           manager.log.info('    %s', line)
    
    return sendmail(job, subject, msg_body, attachments, manager.log)


# ---------------------------------------------------------------------------
def param_encode(lst, encoding):
    """return a new encoded list of string"""
    if isinstance(lst, list):
        newlst=lst[:]
        for i, v in enumerate(lst):
            newlst[i]=v.encode(encoding)
        return newlst
    elif isinstance(lst, unicode):
        return lst.encode(encoding)
    else:
        return lst

#======================================================================
#
# Destinations
#
#======================================================================
class DestinationSyntaxError(Exception):
    pass

class Destinations:
    
    variables={} # default value for extra variable in Destination 
    types={}    # allowed types in Destination

    def __init__(self, raw, archiver):
        
        self.destinations=[]
        st=raw
        self.full_cycle=7
        self.archiver=archiver
        self.used_types=set()
        
        while st:
            if st[0]!='<':
                raise DestinationSyntaxError, 'starting "<" missing'
            
            try:
                pos=st.index('>')
            except ValueError:
                raise DestinationSyntaxError, 'a ">" is missing'
            
            selector=st[1:pos]
            st=st[pos+1:]
        
            pos=st.find('<')
            if pos>=0:
                target=st[:pos]
                st=st[pos:]
            else:
                target=st
                st=''
            
            typ, period=selector.split('=', 1)

            try:
                if typ.lower()=='none':
                    typ='none'
                else:
                    typ=self.archiver.types[typ.lower()]
            except KeyError:
                raise DestinationSyntaxError, 'type "%s" unknown' % (typ, )

            self.used_types.add(typ.lower())

            #print "SELECTOR<%s=%s>%s" % (typ,period,target )
            #print "CARRY=%s" % st
            period=cron.Cron(period)
            if period.weekdivisor!=None:
                self.full_cycle=max(self.full_cycle, period.weekdivisor*7)
            if period.months:
                self.full_cycle=max(self.full_cycle, 63)
                
            self.destinations.append((typ, period, target))


    def match(self, day, night_shift, variables={}):
        #print '---------------------------------', variables
        night_day=day
        if day.hour<12:
            night_day=night_day-timedelta(days=1)
        if night_shift:
            backup_day=night_day
        else:
            backup_day=day
            
        globals=dict(os.environ)
        globals.update(dict(nday='%02d'%(night_day.day,), nmonth='%02d'%(night_day.month,), nyear=str(night_day.year,), nweekday=night_day.weekday(), nyearday='%03d'%(night_day.timetuple().tm_yday,),))
        globals['nweekdayname']=night_day.strftime('%a')
        globals['nmonthname']=night_day.strftime('%b')
        
        epoch=calendar.timegm(backup_day.timetuple())
        globals['epoch']=str(int(epoch))
        globals['week']='0'
        globals['nweek']='0'
        globals['nmonthweek']=night_day.day/7
        
        globals.update(self.archiver.variables)
        globals.update(variables)

        for typ, period, target in self.destinations:
            #print 'ASX typ, per, tgt', typ, period, target
            if period:
                #print 'Month', backup_day.month-1, period.months
                if period.months and not backup_day.month-1 in period.months:
                    continue
                #print 'Day of week', backup_day.isoweekday()-1, period.daysofweek
                if period.daysofweek and not backup_day.isoweekday()-1 in period.daysofweek:
                    continue
                #print 'Day of month', backup_day.day-1, period.daysofmonth
                if period.daysofmonth and not backup_day.day-1 in period.daysofmonth:
                    continue
                if period.firstdayofweek!=None:
                    nday=int(epoch/86400)-period.firstdayofweek+3
                    week=int(nday/7)%period.weekdivisor
                    if period.weekselector!=None and week!=period.weekselector:
                        continue
                    globals['week']=week
                    # The night day
                    nday=int(calendar.timegm(night_day.timetuple())/86400)-period.firstdayofweek+3
                    week=int(nday/7)%period.weekdivisor
                    if period.weekselector!=None and week!=period.weekselector:
                        continue
                    globals['nweek']=week

            #print 'MATCH', typ, target
            rawtarget=target
            for exp in re.findall('\{([^}]*)\}', target):
                try:
                    target=target.replace('${%s}' % (exp,), str(eval(exp, globals, {})))
                except Exception, e:
                    raise DestinationSyntaxError, 'Invalid expression: "%s"' % (exp, )
            target=target.encode('utf-8')
            try:
                target=backup_day.strftime(target)
            except:
                raise DestinationSyntaxError, 'Invalid expression: "%s"' % (target, )
            target=target.decode('utf-8')

            return typ, target
            
        return None, None


# ---------------------------------------------------------------------------

class MyMemoryHandler(logging.handlers.MemoryHandler):

     def shouldFlush(self, record):
        """
        Check for buffer full and drop oldest records.
        """
        if len(self.buffer) >= self.capacity:
            self.buffer=self.buffer[-self.capacity:]
            
        return False

# -------------------------------------------------------------------------
def ErrorDecode(exception, encoding='utf-8'):
    """This is the equivalent of my windows.WindowsErrorDecode but for 
    non windows OS"""
    st=str(exception)
    try:
        return st.decode(encoding)
    except UnicodeDecodeError:
        return st.decode('ascii', 'replace')

# ---------------------------------------------------------------------------
class Manager:
    
    def __init__(self, hostname, log):
        self.hostname=hostname
        self.log=log
        self.default_encoding=locale.getdefaultlocale()[1]
        self.console_encoding=sys.stdout.encoding
        self.formater=logging.Formatter('%(asctime)s %(levelname)-3.3s %(message)s', '%Y-%m-%d %H:%M:%S')
        
    def start_logging(self):
        self.string_log=StringIO.StringIO()
        self.stream_log=logging.StreamHandler(self.string_log)
        self.stream_log.setFormatter(self.formater)
        self.log.addHandler(self.stream_log)

    def stop_logging(self):
        self.stream_log.flush()
        self.log.removeHandler(self.stream_log)
        self.stream_log.close()
        log_output=self.string_log.getvalue()
        # self.string_log.close() don't close it !
        return log_output

# ---------------------------------------------------------------------------
class Archiver:
    
    name=None
    exe=None

    types=dict()
    variables=dict()
 
    def __init__(self):
        self.night_shift=True
        self._hidden_password=[]
        
    def hide_password(self, st):
        """replace password by fake text""" 
        for password in self._hidden_password:
            st=st.replace(password, '********')
        return st
            
    def register_password(self, password):
        """register a new password"""
        if password:
            self._hidden_password.append(password)
    
    # -----------------------------------------------------------------------


# =====================================================================
class UnixRegistry:

    directory='.mksbackup'
    
    def __init__(self, master_key):
        path=os.path.join(os.path.expanduser('~'), self.directory)
        if not os.path.isdir(path):
            os.mkdir(path)
        self.filename=os.path.join(path, master_key)
        
    def LoadValues(self, keys=[]):
        values=dict()
        try:
            for line in open(self.filename).read().split('\n'):
                line=line.strip()
                if not line: continue
                key, value=line.split('=', 1)
                values[key]=value
        except IOError:
            pass
        return values    
                
    def SaveValues(self, values):
        # update old values
        old_values=self.LoadValues()
        old_values.update(values)
        data=""
        for k, v in old_values.iteritems():
            data+='%s=%s\n' % (k, str(v))
        filedata=open(self.filename, "w")
        filedata.write(data)
        filedata.close()
        
    def Close(self):
        pass


# Package init file for pywbem

#
# (C) Copyright 2004,2006 Hewlett-Packard Development Company, L.P.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2 of the
# License.
#   
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#   
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Author: Tim Potter <tpot@hp.com>
#         Martin Pool <mbp@hp.com>

"""pywbem - WBEM client bindings for Python.

PyWBEM is a Python library for making CIM operations over HTTP using
the CIM-XML protocol.  It is based on the idea that a good WBEM
client should be easy to use and not necessarily require a large
amount of programming knowlege.  PyWBEM is suitable for a large range
of tasks from simply poking around to writing web and GUI
applications.

"""

# There are submodules, but clients shouldn't need to know about them.
# Importing just this module is enough.

# These are explicitly safe for 'import *'

from cim_types import *
from cim_constants import *
from cim_operations import *
from cim_obj import *
from tupleparse import ParseError




#
# ghettovcb.py
#
# mkbackup ghettovcb frontent
#

import os, string, subprocess, fnmatch, platform
import shutil
from datetime import datetime, timedelta

from archiver import *
import cron


class Tar(Archiver):
    
    name='tar'
    exe='tar'

    types=dict(normal='normal', full='normal', incremental='incremental', inc='incremental', differential='differential', diff='differential', copy='copy', daily='daily')

    def __init__(self):
        Archiver.__init__(self)
        
        self.include=[]
        self.exclude_dir=[]
        self.exclude=[]
        self.find_bin=None
        self.tar_bin=None
        self.snapshot=None
        self.destination=None
        self.index='auto'
        self.index_file=None
        self.error='auto'
        self.error_file=None
        self.error_find='auto'
        self.error_find_file=None
        
        self.tar_options=[]
        self.find_options=[]
        self.gzip_options=[]
        self.bzip2_options=[]
        
        self.find_args=[]
        self.tar_args=[]
        self.tar_env=None
        
        self.target=None
        self.type=None
        self.target_dir=None

    
    def load(self, job, manager):

        log=manager.log
        now=manager.now

        errors, warnings, extra={}, {}, ''

        #
        # check job config
        #

        self.include=job.get('include', [])
        if not self.include:
            errors['include']='option mandatory'
        else:
            self.include=map(string.strip, self.include.split('\n'))
            for path in self.include:
                if not os.path.isdir(path) and not os.path.isfile(path):
                    warnings['include']='file or directory not found'

        self.exclude_dir=job.get('exclude_dir', [])
        if self.exclude_dir:
            self.exclude_dir=map(string.strip, self.exclude_dir.split('\n'))
            for path in self.exclude_dir:
                if not os.path.isdir(path):
                    warnings['exclude_dir']='directory not found'

        self.exclude=job.get('exclude', [])
        if self.exclude:
            self.exclude=map(string.strip, self.exclude.split('\n'))
                    
        self.find_bin=job.get('find', None)
        if self.find_bin:
            if not os.path.isfile(self.find_bin):
                errors['find']='file not found'
            elif not os.access(self.find_bin, os.X_OK):
                errors['find']='file cannot be executed'

        self.tar_bin=job.get('tar', None)
        if self.tar_bin:
            if not os.path.isfile(self.tar_bin):
                errors['tar']='file not found'
            elif not os.access(self.tar_bin, os.X_OK):
                errors['tar']='file cannot be executed'

        self.snapshot=job.get('snapshot', None)
        if self.snapshot:
            if not os.path.isdir(os.path.dirname(self.snapshot)):
                try:
                    os.makedir(os.path.dirname(self.snapshot))
                except:
                    errors['snapshot']='directory not found: '+os.path.dirname(self.snapshot)

        self.destination=job.get('destination', None)
        if not self.destination:
            errors['destination']='option mandatory'
        else:
            self.destination=self.destination.replace('\n','')
            try:
                self.destination=Destinations(self.destination, self)
            except (DestinationSyntaxError, cron.CronException), e:
                errors['destination']='syntax error: %s' % (str(e), )
            else:
                self.type, self.target=self.destination.match(now, self.night_shift)
                if ('differential' in self.destination.used_types or 'incremental' in self.destination.used_types) and not self.snapshot:
                    errors['destination']='parameter "snapshot" required to do differential or incremental backups'
                     
                if self.type!=None and self.type!='none':
                    self.target_dir=os.path.dirname(self.target)
                    if not os.path.isdir(self.target_dir):
                        try:
                            os.makedirs(self.target_dir)
                        except OSError, e:
                            errors['destination']='directory not found: %s (%s)' % (self.target_dir, e)

        for attr in ('index', 'error', 'error_find'):
            filename=job.get(attr, 'auto')
            setattr(self, attr, filename)
            if not filename in ('auto', 'no'):
                head, tail=os.path.split(filename)
                if not os.path.isfile(filename) or not os.access(filename, os.W_OK):
                    if not os.path.isdir(head):
                        errors[attr]='directory not found'
                    elif not os.access(head, os.W_OK):
                        errors[attr]='directory not writable'
                    
        if not errors:
            for name in ('tar_options', 'find_options', 'gzip_options', 'bzip2_options'):
                option=job.get(name, None)
                if not option:
                    continue
                try:
                    options=quoted_string_list(option)
                except ValueError:
                    errors[name]='not a valid quoted string list'
                else:
                    setattr(self, name, options)

            if self.type!=None and self.type!='none':

                # find_args
                self.find_args=[ 'find' ] + self.include
                for path in self.exclude_dir:
                    self.find_args+=[ "-path" , path, "-prune", "-o" ]
                self.find_args+=[ "-not", "-type", "s" ] # exclude socket
                self.find_args+=self.find_options
                self.find_args.append("-print")
        
                if not self.find_bin:
                    for self.find_bin in [ '/usr/local/bin/find', '/bin/find', '/usr/bin/find', 'find' ]:
                        if os.access(self.find_bin, os.X_OK):
                            break
        
                extra+='find_cmdline=%s\n' % '  '.join(self.find_args)
                extra+='find_bin=%s\n' % self.find_bin
                
                # tar_args
                self.tar_args=[ 'tar', 'cvTf', '-', self.target, '--no-recursion' ] + self.tar_options
                if self.snapshot:
                    if self.type=='normal':
                        self.tar_args.append('--listed-incremental='+self.snapshot)
                    elif self.type in ('differential', 'incremental'):
                        self.tar_args.append('--listed-incremental='+self.snapshot+'-1')
        
                self.tar_env=None
                if self.gzip_options:
                    self.tar_env=os.environ
                    self.tar_env['GZIP']=' '.join(self.gzip_options)
                if self.bzip2_options:
                    if not self.tar_env:
                        self.tar_env=os.environ
                    self.tar_env['BZIP2']=' '.join(self.bzip2_options)
        
                if not self.tar_bin:
                    for self.tar_bin in [ '/usr/local/bin/tar', '/bin/tar', '/usr/bin/tar', 'tar' ]:
                        if os.access(self.tar_bin, os.X_OK):
                            break
        
                extra+='tar_cmdline=%s\n' % '  '.join(self.tar_args)
                extra+='tar_bin=%s\n' % self.tar_bin
        
        return errors, warnings, extra
    
    def run(self, command, job, manager):

        log=manager.log
        now=manager.now

        target_noext=self.target
        for ext in ('.tar', '.tar.gz', '.tgz', '.tbz2', '.tbz'):
            if target_noext.endswith(ext):
                target_noext=target_noext[:-len(ext)]
                break
            
        if self.index=='auto':
            self.index_file=target_noext+'.idx'
        elif self.index=='no':
            self.index_file='/dev/null'
        else:
            self.index_file=self.index

        if self.error=='auto':
            self.error_file=target_noext+'.err'
        elif self.error=='no':
            self.error_file='/dev/null'
        else:
            self.error_file=self.error

        if self.error_find=='auto':
            self.error_find_file=target_noext+'_find.err'
        elif self.error_find=='no':
            self.error_find_file='/dev/null'
        else:
            self.error_find_file=self.error_find

        if self.snapshot:
            if self.type=='normal':
                try:
                    os.unlink(self.snapshot)
                except:
                    pass
                try:
                    os.unlink(self.snapshot+'-1')
                except:
                    pass
            elif self.type=='differential':
                if os.path.isfile(self.snapshot):
                    shutil.copyfile(self.snapshot, self.snapshot+'-1')
            elif self.type=='incremental':
                if not os.path.isfile(self.snapshot+'-1'):
                    if os.path.isfile(self.snapshot):
                        shutil.copyfile(self.snapshot, self.snapshot+'-1')
        
        start=int(time.time())
        p_find = subprocess.Popen(self.find_args, executable=self.find_bin, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=os.open(self.error_find_file, os.O_WRONLY|os.O_CREAT|os.O_TRUNC))
        p_tar = subprocess.Popen(self.tar_args, executable=self.tar_bin, env=self.tar_env, stdin=subprocess.PIPE, stdout=os.open(self.index_file, os.O_WRONLY|os.O_CREAT|os.O_TRUNC), stderr=os.open(self.error_file, os.O_WRONLY|os.O_CREAT|os.O_TRUNC))

        p_find.stdin.close()
        file_count, file_excluded=0, 0
        #output = p_tar.communicate()[0]
        for line in p_find.stdout:
            filename=line.rstrip('\n')
            match=0
            for pattern in self.exclude:
                match=fnmatch.fnmatchcase(filename, pattern)
                if match:
                    file_excluded+=1
                    log.debug('exclude file: %s', filename)
                    break
            if not match:
                p_tar.stdin.write(line)
                file_count+=1
                
        p_tar.stdin.close()   
        
        p_find.wait()
        # print 'returncode find=%r' % (p_find.returncode, )

        p_tar.wait()
        
        end=int(time.time())

        # print 'returncode find=%r, tar=%r' % (p_find.returncode, p_tar.returncode)
        
        if p_find.returncode!=0:
            backup_status='ERR'
        else:
            if p_tar.returncode==0:
                backup_status='OK'
            elif p_tar.returncode==1:
                backup_status='WAR'
            else:
                backup_status='ERR'

        status=u''
        status+='name=%s\r\n' % job['name']
        status+='program=%s\r\n' % self.name
        status+='version=%s\r\n' % self.__version__
        status+='hostname=%s\r\n' % platform.node()

        status+='status=%s\r\n' % backup_status
        
        status+='tar_exit_code=%r\r\n' % p_tar.returncode
        status+='find_exit_code=%r\r\n' % p_find.returncode
        status+='start=%s\r\n' % time.ctime(start)
        status+='end=%s\r\n' % time.ctime(end)
        
        status+='file_count=%d\r\n' % file_count
        status+='file_excluded=%d\r\n' % file_excluded
        
        #
        # target file and directory
        #
        target_dir_list, _total_size=list_dir(self.target_dir, log)

        status+='target=%s\r\n' % self.target
        try:
            stat=os.stat(self.target)
        except Exception, e:
            log.error('checking target file: %s', e)
            status+='target_stat_err=%s\r\n' % e
        else:
            status+='target_size=%d\r\n' % stat.st_size
            status+='target_mtime_epoch=%d\r\n' % stat.st_mtime
            status+='target_mtime=%s\r\n' % time.ctime(stat.st_mtime)
        
        try:
            total_free_bytes=free_space(self.target_dir, log)
        except Exception, e:
            log.error('checking target directory: %s', e)
            status+='target_free_space=%s\r\n' % e
        else:
            status+='target_free_space=%d\r\n' % total_free_bytes
        
        status+='start_epoch=%d\r\n' % start
        status+='end_epoch=%d\r\n' % end
        status+='cmd_tar=%s\r\n' % ' '.join(self.tar_args)
        status+='cmd_find=%s\r\n' % ' '.join(self.find_args)

        attachments=[ # the type, subtype and coding is not used
                       ('dir.txt', None, target_dir_list.encode('utf-8'), 'text', 'plain', 'utf-8'),
                    ]

        if self.error!='no':
            attachments.append(('err.txt', self.error_file, None, 'text', 'plain', 'utf-8'))
            
        if self.error_find!='no':
            attachments.append(('err_find.txt', self.error_find_file, None, 'text', 'plain', 'utf-8'))


        return backup_status, status, attachments

